# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("argilla/magpie-ultra-v0.1")
class ArgillaMagpieUltraDataset(BaseSftDataset):
    """Dataset class for the argilla/magpie-ultra-v0.1 dataset."""

    default_dataset = "argilla/magpie-ultra-v0.1"

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Transform a dataset example into a Conversation object."""
        instruction: str = example.get("instruction", None) or ""
        response: str = example.get("response", None) or ""

        messages = [
            Message(role=Role.USER, content=instruction),
            Message(role=Role.ASSISTANT, content=response),
        ]

        return Conversation(messages=messages)


@register_dataset("Magpie-Align/Llama-3-Magpie-Pro-1M-v0.1")
@register_dataset("Magpie-Align/Magpie-Pro-300K-Filtered")
class MagpieProDataset(BaseSftDataset):
    """Dataset class for the Magpie-Align/Llama-3-Magpie-Pro-1M-v0.1 dataset."""

    default_dataset = "Magpie-Align/Llama-3-Magpie-Pro-1M-v0.1"

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Transform a dataset example into a Conversation object."""
        conversation = example.get("conversations")

        if conversation is None:
            raise ValueError("Conversation is None")

        messages = []
        for message in conversation:
            if message["from"] == "human":
                role = Role.USER
            elif message["from"] == "gpt":
                role = Role.ASSISTANT
            else:
                raise ValueError(f"Unknown role: {message['from']}")
            content = message.get("value", "")
            messages.append(Message(role=role, content=content))
        return Conversation(messages=messages)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Optional, cast

import peft
import transformers

from oumi.core.configs import TrainingConfig
from oumi.core.configs.params.peft_params import PeftSaveMode
from oumi.core.distributed import is_world_process_zero
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.trainers.base_trainer import BaseTrainer
from oumi.utils.logging import logger


class HuggingFaceTrainer(BaseTrainer):
    def __init__(
        self,
        hf_trainer: transformers.Trainer,
        processor: Optional[BaseProcessor] = None,
    ):
        """Initializes HuggingFace-specific Trainer version."""
        self._hf_trainer = hf_trainer
        self._processor = processor

    def train(self, resume_from_checkpoint: Optional[str] = None) -> None:
        """Trains a model."""
        self._hf_trainer.train(resume_from_checkpoint=resume_from_checkpoint)

    def save_state(self) -> None:
        """See base class.

        Saves the Trainer state, since Trainer.save_model saves only the tokenizer
        with the model.

        HuggingFace normally writes state into "trainer_state.json" under output_dir.
        """
        if not is_world_process_zero():
            return

        self._hf_trainer.save_state()

    def save_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model's weights to the specified output directory.

        Args:
            config: The Oumi training config.
            final: Whether this is the final model being saved during training.
                - Applies optimizations for the final model checkpoint.
                - In the case of FSDP, this will always save the FULL_STATE_DICT
                instead of the default STATE_DICT.

        Returns:
            None
        """
        if self._hf_trainer.is_fsdp_enabled:
            # FSDP is enabled, so we need to save the model in a special way.
            return self._save_fsdp_model(config=config, final=final)
        else:
            return self._save_model(config=config, final=final)

    def _save_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model's weights to the specified output directory."""
        if not is_world_process_zero():
            return

        output_dir = config.training.output_dir
        if not config.training.use_peft:
            self._hf_trainer.save_model(output_dir)
        else:
            if config.peft.peft_save_mode == PeftSaveMode.MERGED:
                # Saving the merged model only saves the model weights, not the
                # tokenizer files and training args. To ensure we're saving all relevant
                # files, we save the PEFT model first, delete the adapter files, then
                # save the merged model.
                # The adapter files are moved to the "adapter/" subdirectory to not
                # interfere with the other saved model files.

                self._hf_trainer.save_model(output_dir)
                output_dir_path = pathlib.Path(output_dir)
                adapter_dir = output_dir_path / "adapter"
                adapter_dir.mkdir(parents=True, exist_ok=True)
                for filename in ["adapter_config.json", "adapter_model.safetensors"]:
                    file_path = output_dir_path / filename
                    if file_path.exists():
                        file_path.rename(adapter_dir / filename)
                    else:
                        logger.warning(
                            f"{filename} not found in {output_dir} when "
                            "attempting to delete during model saving."
                        )

                model = cast(peft.LoraModel, self._hf_trainer.model)
                merged_model = model.merge_and_unload(progressbar=True, safe_merge=True)
                merged_model = cast(transformers.PreTrainedModel, merged_model)
                merged_model.save_pretrained(output_dir)
            elif config.peft.peft_save_mode == PeftSaveMode.ADAPTER_ONLY:
                # Save the LoRA adapter (doesn't include the base model).
                self._hf_trainer.save_model(output_dir)
            elif config.peft.peft_save_mode == PeftSaveMode.ADAPTER_AND_BASE_MODEL:
                self._hf_trainer.save_model(output_dir)
                # Saving the base model requires a separate call.
                assert self._hf_trainer.model is not None, (
                    "Model should not be None when using PEFT"
                )
                model = cast(
                    transformers.PreTrainedModel, self._hf_trainer.model.base_model
                )
                model.save_pretrained(output_dir)
            else:
                raise ValueError(
                    f"Unsupported PEFT save mode: {config.peft.peft_save_mode}"
                )
        logger.info(f"Model has been saved at {output_dir}")

        if self._processor is not None:
            self._processor.save_config(output_dir)
            logger.info(f"Processor config has been saved at {output_dir}")

    def _save_fsdp_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model's weights to the specified output directory.

        For FSDP, all ranks should call into this function
        """
        if final:
            # For the final checkpoint, we need to save the FULL_STATE_DICT instead of
            # the default STATE_DICT.
            if (
                self._hf_trainer.is_fsdp_enabled
                and self._hf_trainer.accelerator.state.fsdp_plugin is not None
            ):
                logger.info("Saving FULL_STATE_DICT for final model checkpoint.")
                self._hf_trainer.accelerator.state.fsdp_plugin.set_state_dict_type(
                    "FULL_STATE_DICT"
                )

        output_dir = config.training.output_dir
        self._hf_trainer.save_model(output_dir)
        logger.info(f"Model has been saved at {output_dir}")

        if self._processor is not None:
            self._processor.save_config(output_dir)
            logger.info(f"Processor config has been saved at {output_dir}")

    def get_last_eval_metrics(self) -> dict:
        """Gets the last evaluation metrics from the trainer.

        Returns:
            A dictionary containing the last evaluation metrics.
        """
        return self._hf_trainer.evaluate()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Annotated, Final, Optional

import typer
from rich.table import Table

import oumi.cli.cli_utils as cli_utils
from oumi.cli.alias import AliasType, try_get_config_name_for_alias
from oumi.utils.logging import logger

_DEFAULT_CLI_PDF_DPI: Final[int] = 200


def infer(
    ctx: typer.Context,
    config: Annotated[
        str,
        typer.Option(
            *cli_utils.CONFIG_FLAGS,
            help="Path to the configuration file for inference.",
        ),
    ],
    interactive: Annotated[
        bool,
        typer.Option("-i", "--interactive", help="Run in an interactive session."),
    ] = False,
    image: Annotated[
        Optional[str],
        typer.Option(
            "--image",
            help=(
                "File path or URL of an input image to be used with image+text VLLMs. "
                "Only used in interactive mode."
            ),
        ),
    ] = None,
    system_prompt: Annotated[
        Optional[str],
        typer.Option(
            "--system-prompt",
            help=(
                "System prompt for task-specific instructions. "
                "Only used in interactive mode."
            ),
        ),
    ] = None,
    level: cli_utils.LOG_LEVEL_TYPE = None,
    verbose: cli_utils.VERBOSE_TYPE = False,
):
    """Run inference on a model.

    If `input_filepath` is provided in the configuration file, inference will run on
    those input examples. Otherwise, inference will run interactively with user-provided
    inputs.

    Args:
        ctx: The Typer context object.
        config: Path to the configuration file for inference.
        output_dir: Directory to save configs
        (defaults to OUMI_DIR env var or ~/.oumi/fetch).
        interactive: Whether to run in an interactive session.
        image: Path to the input image for `image+text` VLLMs.
        system_prompt: System prompt for task-specific instructions.
        level: The logging level for the specified command.
        verbose: Enable verbose logging with additional debug information.
    """
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            try_get_config_name_for_alias(config, AliasType.INFER),
        )
    )

    # Delayed imports
    from oumi import infer as oumi_infer
    from oumi import infer_interactive as oumi_infer_interactive
    from oumi.core.configs import InferenceConfig
    from oumi.utils.image_utils import (
        create_png_bytes_from_image_list,
        load_image_png_bytes_from_path,
        load_image_png_bytes_from_url,
        load_pdf_pages_from_path,
        load_pdf_pages_from_url,
    )
    # End imports

    parsed_config: InferenceConfig = InferenceConfig.from_yaml_and_arg_list(
        config, extra_args, logger=logger
    )
    parsed_config.finalize_and_validate()

    if verbose:
        # Print configuration for verification
        parsed_config.print_config(logger)

    # https://stackoverflow.com/questions/62691279/how-to-disable-tokenizers-parallelism-true-false-warning
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    input_image_png_bytes: Optional[list[bytes]] = None
    if image:
        image_lower = image.lower()
        if image_lower.startswith("http://") or image_lower.startswith("https://"):
            if image_lower.endswith(".pdf"):
                input_image_png_bytes = create_png_bytes_from_image_list(
                    load_pdf_pages_from_url(image, dpi=_DEFAULT_CLI_PDF_DPI)
                )
            else:
                input_image_png_bytes = [load_image_png_bytes_from_url(image)]
        else:
            if image_lower.endswith(".pdf"):
                input_image_png_bytes = create_png_bytes_from_image_list(
                    load_pdf_pages_from_path(image, dpi=_DEFAULT_CLI_PDF_DPI)
                )
            else:
                input_image_png_bytes = [load_image_png_bytes_from_path(image)]
    if parsed_config.input_path:
        if interactive:
            logger.warning(
                "Input path provided, skipping interactive inference. "
                "To run in interactive mode, do not provide an input path."
            )
        generations = oumi_infer(parsed_config)
        # Don't print results if output_filepath is provided.
        if parsed_config.output_path:
            return
        table = Table(
            title="Inference Results",
            title_style="bold magenta",
            show_edge=False,
            show_lines=True,
        )
        table.add_column("Conversation", style="green")
        for generation in generations:
            table.add_row(repr(generation))
        cli_utils.CONSOLE.print(table)
        return
    if not interactive:
        logger.warning(
            "No input path provided, running in interactive mode. "
            "To run with an input path, provide one in the configuration file."
        )
    return oumi_infer_interactive(
        parsed_config,
        input_image_bytes=input_image_png_bytes,
        system_prompt=system_prompt,
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from typing_extensions import override

from oumi.core.datasets.base_grpo_dataset import BaseExperimentalGrpoDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation

_SYSTEM_PROMPT = (
    "Your final answer should be an integer written as digits and formatted as "
    r'"\boxed{your_answer}". For example, if the answer is 42, '
    r'you should output "\boxed{42}".'
)


@register_dataset("oumi-ai/oumi-letter-count-clean")
@register_dataset("oumi-ai/oumi-letter-count")
class LetterCountGrpoDataset(BaseExperimentalGrpoDataset):
    """Dataset class for the `oumi-ai/oumi-letter-count` dataset.

    A sample from the dataset::

        {
            "conversation_id": "oumi_letter_count_0",
            "messages": [
                {
                    "content": "Can you let me know how many 'r's are in 'pandered'?",
                    "role": "user",
                }
            ],
            "metadata": {
                "letter": "r",
                "letter_count_integer": 1,
                "letter_count_string": "one",
                "unformatted_prompt": "Can you let me know how many {letter}s are in {word}?",
                "word": "pandered",
            },
        }

    """  # noqa: E501

    default_dataset = "oumi-ai/oumi-letter-count"

    @override
    def transform(self, sample: pd.Series) -> dict:
        """Validate and transform the sample into Python `dict`."""
        # Add system prompt before user prompt.
        system_message = {"content": _SYSTEM_PROMPT, "role": "system"}
        messages = [system_message, sample["messages"][0]]
        return {
            "prompt": messages,
            "letter_count": sample["metadata"]["letter_count_integer"],
        }

    @override
    def transform_conversation(self, sample: pd.Series) -> Conversation:
        """Converts the input sample to a Conversation.

        Args:
            sample (dict): The input example.

        Returns:
            Conversation: The resulting conversation.

        """
        # Example is already in conversation format and only needs light processing.
        sample_dict = sample.to_dict()
        # Add system prompt before user prompt.
        system_message = {"content": _SYSTEM_PROMPT, "role": "system"}
        messages = [system_message, sample["messages"][0]]
        sample_dict["messages"] = messages
        return Conversation.from_dict(sample_dict)



from unittest.mock import patch

import pytest

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.sambanova_inference_engine import SambanovaInferenceEngine


@pytest.fixture
def sambanova_engine():
    return SambanovaInferenceEngine(
        model_params=ModelParams(model_name="Meta-Llama-3.1-8B-Instruct"),
        remote_params=RemoteParams(api_key="test_api_key", api_url="<placeholder>"),
    )


def test_convert_conversation_to_api_input(sambanova_engine):
    """Test conversion of conversation to SambaNova API input format."""
    conversation = Conversation(
        messages=[
            Message(content="System message", role=Role.SYSTEM),
            Message(content="User message", role=Role.USER),
            Message(content="Assistant message", role=Role.ASSISTANT),
        ]
    )
    generation_params = GenerationParams(
        max_new_tokens=100,
        temperature=0.7,
        top_p=0.9,
        stop_strings=["stop"],
    )

    result = sambanova_engine._convert_conversation_to_api_input(
        conversation, generation_params, sambanova_engine._model_params
    )

    # Verify the API input format
    assert result["model"] == "Meta-Llama-3.1-8B-Instruct"
    assert len(result["messages"]) == 3
    assert result["messages"][0]["content"] == "System message"
    assert result["messages"][0]["role"] == "system"
    assert result["messages"][1]["content"] == "User message"
    assert result["messages"][1]["role"] == "user"
    assert result["messages"][2]["content"] == "Assistant message"
    assert result["messages"][2]["role"] == "assistant"
    assert result["max_tokens"] == 100
    assert result["temperature"] == 0.7
    assert result["top_p"] == 0.9
    assert result["stop"] == ["stop"]
    assert result["stream"] is False


def test_convert_api_output_to_conversation(sambanova_engine):
    """Test conversion of SambaNova API output to conversation."""
    original_conversation = Conversation(
        messages=[
            Message(content="User message", role=Role.USER),
        ],
        metadata={"key": "value"},
        conversation_id="test_id",
    )
    api_response = {
        "choices": [{"message": {"content": "Assistant response", "role": "assistant"}}]
    }

    result = sambanova_engine._convert_api_output_to_conversation(
        api_response, original_conversation
    )

    assert len(result.messages) == 2
    assert result.messages[0].content == "User message"
    assert result.messages[1].content == "Assistant response"
    assert result.messages[1].role == Role.ASSISTANT
    assert result.metadata == {"key": "value"}
    assert result.conversation_id == "test_id"


def test_convert_api_output_to_conversation_error_handling(sambanova_engine):
    """Test error handling in API output conversion."""
    original_conversation = Conversation(
        messages=[Message(content="User message", role=Role.USER)]
    )

    # Test empty choices
    with pytest.raises(RuntimeError, match="No choices found in API response"):
        sambanova_engine._convert_api_output_to_conversation(
            {"choices": []}, original_conversation
        )

    # Test missing message
    with pytest.raises(RuntimeError, match="No message found in API response"):
        sambanova_engine._convert_api_output_to_conversation(
            {"choices": [{}]}, original_conversation
        )


def test_get_request_headers(sambanova_engine):
    """Test generation of request headers."""
    remote_params = RemoteParams(api_key="test_api_key", api_url="<placeholder>")

    with patch.object(
        SambanovaInferenceEngine,
        "_get_api_key",
        return_value="test_api_key",
    ):
        result = sambanova_engine._get_request_headers(remote_params)

    assert result["Content-Type"] == "application/json"
    assert result["Authorization"] == "Bearer test_api_key"


def test_get_supported_params(sambanova_engine):
    """Test supported generation parameters."""
    supported_params = sambanova_engine.get_supported_params()

    assert supported_params == {
        "max_new_tokens",
        "stop_strings",
        "temperature",
        "top_p",
    }



"""Minimal multi-modal training script with CLI arguments and custom collator.

Run the script using:
   python scripts/benchmarks/minimal_multimodal_training.py \
    --model-name<model_name> --dataset-name <dataset_name>

For multi-GPU training, use torchrun:
   torchrun --standalone --nproc-per-node=$(nvidia-smi --list-gpus | wc -l) \
        scripts/benchmarks/minimal_multimodal_training.py \
            --model-name <model_name> --dataset-name <dataset_name>

Working configs:
    --model-name Salesforce/blip2-opt-2.7b --dataset-name merve/vqav2-small
    --model-name Salesforce/blip2-opt-2.7b --dataset-name nlphuji/flickr30k
    --model-name Qwen/Qwen2-VL-2B-Instruct --dataset-name merve/vqav2-small
    --model-name Qwen/Qwen2-VL-2B-Instruct --dataset-name nlphuji/flickr30k
    --model-name Qwen/Qwen2.5-VL-3B-Instruct --dataset-name merve/vqav2-small
    --model-name llava-hf/llava-1.5-7b-hf --dataset-name merve/vqav2-small --test-fsdp
    --model-name llava-hf/llava-1.5-7b-hf --dataset-name nlphuji/flickr30k --test-fsdp


"""

from enum import Enum
from pprint import pformat
from typing import Final, NamedTuple, Optional

import torch
import typer

import oumi.core.constants as constants
from oumi.builders import (
    build_data_collator,
    build_dataset,
    build_model,
    build_processor,
    build_tokenizer,
)
from oumi.core.configs import (
    FSDPParams,
    ModelParams,
    TrainingParams,
)
from oumi.core.distributed import (
    cleanup_distributed,
    init_distributed,
    is_distributed,
    is_local_process_zero,
)
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.trainers.oumi_trainer import Trainer
from oumi.utils.str_utils import sanitize_run_name
from oumi.utils.torch_utils import (
    log_model_summary,
)


class ModelName(str, Enum):
    LLAVA = "llava-hf/llava-1.5-7b-hf"
    BLIP2 = "Salesforce/blip2-opt-2.7b"
    LLAMA_11B_VISION_INSTRUCT = "meta-llama/Llama-3.2-11B-Vision-Instruct"
    QWEN2_VL = "Qwen/Qwen2-VL-2B-Instruct"
    QWEN2_5_3B_VL = "Qwen/Qwen2.5-VL-3B-Instruct"
    CHAMELEON = "facebook/chameleon-7b"
    PALIGEMMA = "google/paligemma-3b-mix-224"
    PHI3_VISION = "microsoft/Phi-3-vision-128k-instruct"  # requires flash-attn
    MOLMOE_1B = "allenai/MolmoE-1B-0924"
    SMOLVLM = "HuggingFaceTB/SmolVLM-Instruct"


_DEFAULT_MLLM_CHAT_TEMPLATE: Final[str] = "llava"
_DEFAULT_MLLM_COLLATOR: Final[str] = "vision_language_with_padding"


class ModelInfo(NamedTuple):
    chat_template: str
    freeze_layers: list[str]
    supports_multiple_images: bool = False
    collator_name: str = _DEFAULT_MLLM_COLLATOR


_MODELS_MAP: dict[ModelName, ModelInfo] = {
    ModelName.BLIP2: ModelInfo(
        chat_template="default",
        freeze_layers=["vision_model"],
    ),
    ModelName.LLAVA: ModelInfo(
        chat_template=_DEFAULT_MLLM_CHAT_TEMPLATE,
        freeze_layers=["vision_tower"],
    ),
    ModelName.QWEN2_VL: ModelInfo(
        chat_template="qwen2-vl-instruct",
        freeze_layers=["visual"],
        collator_name="vision_language_sft",
    ),
    ModelName.QWEN2_5_3B_VL: ModelInfo(
        chat_template="qwen2-vl-instruct",
        freeze_layers=["visual"],
    ),
    ModelName.CHAMELEON: ModelInfo(
        chat_template=_DEFAULT_MLLM_CHAT_TEMPLATE,
        freeze_layers=["model.vqmodel"],
    ),
    ModelName.PALIGEMMA: ModelInfo(
        chat_template=_DEFAULT_MLLM_CHAT_TEMPLATE,
        freeze_layers=["vision_tower"],
    ),
    ModelName.PHI3_VISION: ModelInfo(
        chat_template="phi3-instruct",
        freeze_layers=["model.vision_embed_tokens"],
    ),
    ModelName.LLAMA_11B_VISION_INSTRUCT: ModelInfo(
        chat_template="llama3-instruct",
        freeze_layers=["vision_model"],
        supports_multiple_images=True,
    ),
    ModelName.MOLMOE_1B: ModelInfo(
        chat_template=_DEFAULT_MLLM_CHAT_TEMPLATE,
        freeze_layers=["model.vision_backbone"],
    ),
    ModelName.SMOLVLM: ModelInfo(
        chat_template=_DEFAULT_MLLM_CHAT_TEMPLATE,
        freeze_layers=["vision_model"],
    ),
}


def _get_freeze_layers(model_name: ModelName) -> list[str]:
    result = []
    if model_name in _MODELS_MAP:
        result = _MODELS_MAP[model_name].freeze_layers
        print(f"Frozen layers: {result}")
    else:
        print(f"No frozen layers defined for {model_name}!")
    return result


def _get_chat_template(model_name: ModelName) -> str:
    result = ""
    if model_name in _MODELS_MAP:
        result = _MODELS_MAP[model_name].chat_template
        print(f"Chat template: {result}")
    else:
        result = _DEFAULT_MLLM_CHAT_TEMPLATE
        print(f"No chat template defined for {model_name}! Defaulting to '{result}'...")
    return result or _DEFAULT_MLLM_CHAT_TEMPLATE


def _get_collator_name(model_name: ModelName) -> str:
    result = ""
    if model_name in _MODELS_MAP:
        result = _MODELS_MAP[model_name].collator_name
        print(f"Collator: {result}")
    else:
        result = _DEFAULT_MLLM_COLLATOR
        print(f"No collator defined for {model_name}! Defaulting to '{result}'...")
    return result or _DEFAULT_MLLM_COLLATOR


def _supports_multiple_images(model_name: ModelName) -> bool:
    result = False
    if model_name in _MODELS_MAP:
        result = _MODELS_MAP[model_name].supports_multiple_images
    return result


class DatasetName(str, Enum):
    MERVE_VQAV2_SMALL = "merve/vqav2-small"
    LLAVA_INSTRUCT_MIX_VSFT = "HuggingFaceH4/llava-instruct-mix-vsft"
    FLICKR = "nlphuji/flickr30k"
    COCO = "coco_captions"
    MNIST_SFT = "mnist_sft"
    DOCMATIX = "HuggingFaceM4/Docmatix"


class _DatasetInfo(NamedTuple):
    default_split: str
    default_subset: Optional[str]
    contains_multiple_images: bool


def _get_dataset_info(dataset_name: DatasetName) -> _DatasetInfo:
    default_split = "train"
    if dataset_name in (DatasetName.FLICKR,):
        # The dataset only has "test" split.
        default_split = "test"
    elif dataset_name in (DatasetName.MERVE_VQAV2_SMALL,):
        default_split = "validation"

    contains_multiple_images: bool = False

    default_subset: Optional[str] = None
    if dataset_name in (DatasetName.DOCMATIX,):
        # The only non-giant subset in the dataset
        default_subset = "zero-shot-exp"
        contains_multiple_images = True

    return _DatasetInfo(
        default_split=default_split,
        default_subset=default_subset,
        contains_multiple_images=contains_multiple_images,
    )


def test_multimodal_trainer(
    model_name: ModelName = ModelName.BLIP2,
    dataset_name: DatasetName = DatasetName.COCO,
    batch_size: int = 2,
    max_steps: int = 20,
    optimizer: str = "sgd",
    logging_steps: int = 5,
    split: Optional[str] = None,
    subset: Optional[str] = None,
    test_inference: bool = False,
    test_save_state: bool = False,
    test_fsdp: bool = False,
):
    """Minimal multi-modal training loop."""
    if is_distributed():
        print("Initializing distributed process group")
        init_distributed()
    else:
        print("Not initializing distributed process group")

    dataset_info = _get_dataset_info(dataset_name)

    if not split:
        split = dataset_info.default_split
    if not subset:
        subset = dataset_info.default_subset

    if dataset_info.contains_multiple_images and not _supports_multiple_images(
        model_name
    ):
        print(
            f"Using multi-image dataset {dataset_name} "
            f"with model {model_name} that doesn't support multiple images! "
            "Training may FAIL!"
        )

    #
    # Init model, processor, and dataset
    #
    model_params = ModelParams(
        model_name=model_name.value,
        torch_dtype_str="bfloat16",
        trust_remote_code=True,
        chat_template=_get_chat_template(model_name),
        freeze_layers=_get_freeze_layers(model_name),  # TODO: fix freeze + fsdp
    )
    if is_local_process_zero():
        print(f"ModelParams:\n{pformat(model_params)}")

    model = build_model(model_params)
    tokenizer: BaseTokenizer = build_tokenizer(model_params)
    processor: BaseProcessor = build_processor(
        model_name.value, tokenizer, trust_remote_code=True
    )

    collator_name = _get_collator_name(model_name)
    dataset_kwargs = dict(processor=processor, limit=100)
    if collator_name == "vision_language_sft":
        dataset_kwargs["return_conversations"] = True
    dataset = build_dataset(
        dataset_name=str(dataset_name.value),
        tokenizer=tokenizer,
        subset=subset,
        split=split,
        dataset_kwargs=dataset_kwargs,
        trust_remote_code=True,
        use_torchdata=True,
    )

    #
    # Set up training parameters
    #
    fsdp_params = FSDPParams(
        enable_fsdp=is_distributed() and test_fsdp, cpu_offload=True
    )

    run_name = sanitize_run_name(
        f"multimodal_test_{model_name.value.split('/')[-1]}_{dataset_name.value}"
    )

    training_params = TrainingParams(
        output_dir=f"output/{run_name}",
        per_device_train_batch_size=batch_size,
        max_steps=max_steps,
        save_steps=0,
        optimizer=(optimizer or "sgd"),
        learning_rate=2e-5,
        warmup_steps=int(max(10, 0.2 * max_steps)),
        max_grad_norm=10,
        lr_scheduler_type="cosine",
        gradient_accumulation_steps=1,
        log_model_summary=False,
        logging_steps=logging_steps,
        include_performance_metrics=True,
    )

    if is_local_process_zero():
        print(f"TrainingParams:\n{pformat(training_params)}")
        if training_params.log_model_summary:
            log_model_summary(model)

    collator_kwargs = {}
    if collator_name == "vision_language_sft":
        collator_kwargs["processor_name"] = model_name.value
        collator_kwargs["processor_kwargs"] = model_params.processor_kwargs
        collator_kwargs["trust_remote_code"] = model_params.trust_remote_code

    # Initialize trainer with custom collator
    collator = build_data_collator(
        collator_name=collator_name,
        tokenizer=tokenizer,
        max_length=model_params.model_max_length,
        label_ignore_index=constants.LABEL_IGNORE_INDEX,
        **collator_kwargs,
    )

    trainer = Trainer(
        model=model,
        processing_class=tokenizer,
        args=training_params,
        train_dataset=dataset,
        fsdp_params=fsdp_params,
        data_collator=collator,
    )

    #
    # Train
    #
    trainer.train()
    if test_save_state:
        trainer.save_state()

    #
    # Test inference
    #
    if test_inference:
        test_input = dataset[0]
        with torch.no_grad():
            output = trainer.model(**test_input)

        print("Test output:", output.keys())
        print("Test output shapes:", {k: v.shape for k, v in output.items()})

    if is_distributed():
        cleanup_distributed()

    print(
        f"Multi-modal training test successful with model {model_name} and "
        f"dataset {dataset_name}!"
    )

    return True


if __name__ == "__main__":
    typer.run(test_multimodal_trainer)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from pprint import pformat
from typing import Callable, Optional, cast

import transformers
import trl

from oumi.core.configs import TrainerType, TrainingParams
from oumi.core.distributed import is_world_process_zero
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.trainers import (
    BaseTrainer,
    HuggingFaceTrainer,
    TrlDpoTrainer,
    VerlGrpoTrainer,
)
from oumi.core.trainers import Trainer as OumiTrainer
from oumi.utils.logging import logger


def build_trainer(
    trainer_type: TrainerType, processor: Optional[BaseProcessor], verbose: bool = False
) -> Callable[..., BaseTrainer]:
    """Builds a trainer creator functor based on the provided configuration.

    Args:
        trainer_type (TrainerType): Enum indicating the type of training.
        processor: An optional processor.
        verbose (bool): Whether to enable verbose logging of training arguments.

    Returns:
        A builder function that can create an appropriate trainer based on the trainer
        type specified in the configuration. All function arguments supplied by caller
        are forwarded to the trainer's constructor.

    Raises:
        NotImplementedError: If the trainer type specified in the
            configuration is not supported.
    """

    def _create_hf_builder_fn(
        cls: type[transformers.Trainer],
    ) -> Callable[..., BaseTrainer]:
        def _init_hf_trainer(*args, **kwargs) -> BaseTrainer:
            training_args = kwargs.pop("args", None)
            training_config = kwargs.pop("training_config", None)
            callbacks = kwargs.pop("callbacks", [])
            if training_args is not None:
                # if set, convert to HuggingFace Trainer args format
                training_args = cast(TrainingParams, training_args)
                training_args.finalize_and_validate()

            hf_args = training_args.to_hf(training_config)
            if verbose and is_world_process_zero():
                logger.info(pformat(hf_args))
            trainer = HuggingFaceTrainer(cls(*args, **kwargs, args=hf_args), processor)
            if callbacks:
                # TODO(OPE-250): Define generalizable callback abstraction
                # Incredibly ugly, but this is the only way to add callbacks that add
                # metrics to wandb. Transformers trainer has no public method of
                # allowing us to control the order callbacks are called.
                training_callbacks = (
                    [transformers.trainer_callback.DefaultFlowCallback]
                    + callbacks
                    # Skip the first callback, which is the DefaultFlowCallback above.
                    + trainer._hf_trainer.callback_handler.callbacks[1:]
                )
                trainer._hf_trainer.callback_handler.callbacks = []
                for c in training_callbacks:
                    trainer._hf_trainer.add_callback(c)
            return trainer

        return _init_hf_trainer

    def _create_oumi_builder_fn() -> Callable[..., BaseTrainer]:
        def _init_oumi_trainer(*args, **kwargs) -> BaseTrainer:
            kwargs_processor = kwargs.get("processor", None)
            if processor is not None:
                if kwargs_processor is None:
                    kwargs["processor"] = processor
                elif id(kwargs_processor) != id(processor):
                    raise ValueError(
                        "Different processor instances passed to Oumi trainer, "
                        "and build_trainer()."
                    )
            return OumiTrainer(*args, **kwargs)

        return _init_oumi_trainer

    def _create_verl_grpo_builder_fn() -> Callable[..., BaseTrainer]:
        def _init_verl_grpo_trainer(*args, **kwargs) -> BaseTrainer:
            return VerlGrpoTrainer(*args, **kwargs)

        return _init_verl_grpo_trainer

    if trainer_type == TrainerType.TRL_SFT:
        return _create_hf_builder_fn(trl.SFTTrainer)
    elif trainer_type == TrainerType.TRL_DPO:
        return _create_hf_builder_fn(TrlDpoTrainer)
    elif trainer_type == TrainerType.TRL_KTO:
        return _create_hf_builder_fn(trl.KTOTrainer)
    elif trainer_type == TrainerType.TRL_GRPO:
        return _create_hf_builder_fn(trl.GRPOTrainer)
    elif trainer_type == TrainerType.TRL_GKD:
        return _create_hf_builder_fn(trl.GKDTrainer)
    elif trainer_type == TrainerType.HF:
        return _create_hf_builder_fn(transformers.Trainer)
    elif trainer_type == TrainerType.OUMI:
        warnings.warn(
            "OUMI trainer is still in alpha mode. "
            "Prefer to use HF trainer when possible."
        )
        return _create_oumi_builder_fn()
    elif trainer_type == TrainerType.VERL_GRPO:
        return _create_verl_grpo_builder_fn()

    raise NotImplementedError(f"Trainer type {trainer_type} not supported.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants and mappings for quantization methods."""

# Supported quantization methods
SUPPORTED_METHODS = [
    # AWQ methods
    "awq_q4_0",
    "awq_q4_1",
    "awq_q8_0",
    "awq_f16",
    # BitsAndBytes methods
    "bnb_4bit",
    "bnb_8bit",
]

# Supported output formats
SUPPORTED_OUTPUT_FORMATS = ["safetensors"]


# Size units for formatting
SIZE_UNITS = ["B", "KB", "MB", "GB", "TB", "PB"]


# Common file extensions for model files
MODEL_FILE_EXTENSIONS = [".safetensors", ".bin", ".pth"]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from collections.abc import Generator, Mapping
from enum import Enum
from types import MappingProxyType
from typing import Any, Callable, Final, NamedTuple, Optional, Union

import pydantic
from jinja2 import Template

import oumi.core.types.proto.generated.conversation_pb2 as pb2


class Role(str, Enum):
    """Role of the entity sending the message."""

    SYSTEM = "system"
    """Represents a system message in the conversation."""

    USER = "user"
    """Represents a user message in the conversation."""

    ASSISTANT = "assistant"
    """Represents an assistant message in the conversation."""

    TOOL = "tool"
    """Represents a tool message in the conversation."""

    def __str__(self) -> str:
        """Return the string representation of the Role enum.

        Returns:
            str: The string value of the Role enum.
        """
        return self.value


_ROLE_TO_PROTO_ROLE_MAP: Final[Mapping[Role, pb2.Role]] = MappingProxyType(
    {
        Role.SYSTEM: pb2.Role.SYSTEM,
        Role.USER: pb2.Role.USER,
        Role.ASSISTANT: pb2.Role.ASSISTANT,
        Role.TOOL: pb2.Role.TOOL,
    }
)
_PROTO_ROLE_TO_ROLE_MAP: Final[Mapping[pb2.Role, Role]] = MappingProxyType(
    {v: k for k, v in _ROLE_TO_PROTO_ROLE_MAP.items()}
)


def _convert_role_to_proto_role(role: Role) -> pb2.Role:
    """Converts a Role enum to Protocol Buffer format."""
    result: pb2.Role = _ROLE_TO_PROTO_ROLE_MAP.get(role, pb2.Role.ROLE_UNSPECIFIED)
    if result == pb2.Role.ROLE_UNSPECIFIED:
        raise ValueError(f"Invalid role: {role}")
    return result


def _convert_proto_role_to_role(role: pb2.Role) -> Role:
    """Converts a Protocol Buffer role format to role."""
    result: Optional[Role] = _PROTO_ROLE_TO_ROLE_MAP.get(role, None)
    if result is None:
        raise ValueError(f"Invalid role: {role}")
    return result


class Type(str, Enum):
    """Type of the message."""

    TEXT = "text"
    """Represents a text message."""

    IMAGE_PATH = "image_path"
    """Represents an image referenced by its file path."""

    IMAGE_URL = "image_url"
    """Represents an image referenced by its URL."""

    IMAGE_BINARY = "image_binary"
    """Represents an image stored as binary data."""

    def __str__(self) -> str:
        """Return the string representation of the Type enum.

        Returns:
            str: The string value of the Type enum.
        """
        return self.value


_CONTENT_ITEM_TYPE_TO_PROTO_TYPE_MAP: Final[Mapping[Type, pb2.ContentPart.Type]] = (
    MappingProxyType(
        {
            Type.TEXT: pb2.ContentPart.TEXT,
            Type.IMAGE_PATH: pb2.ContentPart.IMAGE_PATH,
            Type.IMAGE_URL: pb2.ContentPart.IMAGE_URL,
            Type.IMAGE_BINARY: pb2.ContentPart.IMAGE_BINARY,
        }
    )
)
_CONTENT_ITEM_PROTO_TYPE_TO_TYPE_MAP: Final[Mapping[pb2.ContentPart.Type, Type]] = (
    MappingProxyType({v: k for k, v in _CONTENT_ITEM_TYPE_TO_PROTO_TYPE_MAP.items()})
)


def _convert_type_to_proto_type(content_type: Type) -> pb2.ContentPart.Type:
    """Converts a type enum to Protocol Buffer format."""
    result: pb2.ContentPart.Type = _CONTENT_ITEM_TYPE_TO_PROTO_TYPE_MAP.get(
        content_type, pb2.ContentPart.TYPE_UNSPECIFIED
    )
    if result == pb2.ContentPart.TYPE_UNSPECIFIED:
        raise ValueError(f"Invalid type: {content_type}")
    return result


def _convert_proto_type_to_type(content_type: pb2.ContentPart.Type) -> Type:
    """Converts a Protocol Buffer type format to type."""
    result: Optional[Type] = _CONTENT_ITEM_PROTO_TYPE_TO_TYPE_MAP.get(
        content_type, None
    )
    if result is None:
        raise ValueError(f"Invalid type: {content_type}")
    return result


class ContentItemCounts(NamedTuple):
    """Contains counts of content items in a message by type."""

    total_items: int
    """The total number of content items in a message."""

    text_items: int
    """The number of text content items in a message."""

    image_items: int
    """The number of image content items in a message."""


class ContentItem(pydantic.BaseModel):
    """A sub-part of `Message.content`.

    For example, a multimodal message from `USER` may include
    two `ContentItem`-s: one for text, and another for image.

    Note:
        Either content or binary must be provided when creating an instance.
    """

    model_config = pydantic.ConfigDict(frozen=True)

    type: Type
    """The type of the content (e.g., text, image path, image URL)."""

    content: Optional[str] = None
    """Optional text content of the content item.

    One of content or binary must be provided.
    """

    binary: Optional[bytes] = None
    """Optional binary data for the message content item, used for image data.

    One of content or binary must be provided.

    The field is required for `IMAGE_BINARY`, and can be optionally populated for
    `IMAGE_URL`, `IMAGE_PATH` in which case it must be the loaded bytes of
    the image specified in the `content` field.

    The field must be `None` for `TEXT`.
    """

    def is_image(self) -> bool:
        """Checks if the item contains an image."""
        return self.type in (Type.IMAGE_BINARY, Type.IMAGE_URL, Type.IMAGE_PATH)

    def is_text(self) -> bool:
        """Checks if the item contains text."""
        return self.type == Type.TEXT

    @pydantic.field_serializer("binary")
    def _encode_binary(self, value: Optional[bytes]) -> str:
        """Encode binary value as base64 ASCII string.

        This is needed for compatibility with JSON.
        """
        if value is None or len(value) == 0:
            return ""
        return base64.b64encode(value).decode("ascii")

    @pydantic.field_validator("binary", mode="before")
    def _decode_binary(cls, value: Optional[Union[str, bytes]]) -> Optional[bytes]:
        if value is None:
            return None
        elif isinstance(value, str):
            return base64.b64decode(value)
        return value

    def model_post_init(self, __context) -> None:
        """Post-initialization method for the `ContentItem` model.

        This method is automatically called after the model is initialized.
        Performs additional validation e.g., to ensure that either content or binary
        is provided for the message.

        Raises:
            ValueError: If fields are set to invalid or inconsistent values.
        """
        if self.binary is None and self.content is None:
            raise ValueError(
                "Either content or binary must be provided for the message item "
                f"(Item type: {self.type})."
            )

        if self.is_image():
            if self.type == Type.IMAGE_BINARY and (
                self.binary is None or len(self.binary) == 0
            ):
                raise ValueError(
                    f"No image bytes in message content item (Item type: {self.type})."
                )
            if self.type in (Type.IMAGE_PATH, Type.IMAGE_URL) and (
                self.content is None or len(self.content) == 0
            ):
                raise ValueError(f"Content not provided for {self.type} message item.")
        else:
            if self.binary is not None:
                raise ValueError(
                    f"Binary can only be provided for images (Item type: {self.type})."
                )

    @staticmethod
    def from_proto(item_proto: pb2.ContentPart) -> "ContentItem":
        """Converts a Protocol Buffer to a content item."""
        if item_proto.HasField("blob") and item_proto.blob:
            return ContentItem(
                type=_convert_proto_type_to_type(item_proto.type),
                binary=item_proto.blob.binary_data,
                content=(item_proto.content or None),
            )
        return ContentItem(
            type=_convert_proto_type_to_type(item_proto.type),
            content=item_proto.content,
        )

    def to_proto(self) -> pb2.ContentPart:
        """Converts a content item to Protocol Buffer format."""
        if self.binary is not None and len(self.binary) > 0:
            return pb2.ContentPart(
                type=_convert_type_to_proto_type(self.type),
                blob=pb2.DataBlob(binary_data=self.binary),
                content=(self.content or None),
            )
        return pb2.ContentPart(
            type=_convert_type_to_proto_type(self.type),
            content=(self.content or None),
        )

    def __repr__(self) -> str:
        """Returns a string representation of the message item."""
        return f"{self.content}" if self.is_text() else f"<{self.type.upper()}>"


class Message(pydantic.BaseModel):
    """A message in a conversation.

    This class represents a single message within a conversation, containing
    various attributes such as role, content, identifier.
    """

    model_config = pydantic.ConfigDict(frozen=True)

    id: Optional[str] = None
    """Optional unique identifier for the message.

    This attribute can be used to assign a specific identifier to the message,
    which may be useful for tracking or referencing messages within a conversation.

    Returns:
        Optional[str]: The unique identifier of the message, if set; otherwise None.
    """

    content: Union[str, list[ContentItem]]
    """Content of the message.

    For text messages, `content` can be set to a string value.
    For multimodal messages, `content` should be a list of content items of
    potentially different types e.g., text and image.
    """

    role: Role
    """The role of the entity sending the message (e.g., user, assistant, system)."""

    def model_post_init(self, __context) -> None:
        """Post-initialization method for the Message model.

        This method is automatically called after the model is initialized.
        It performs additional validation to ensure that either content or binary
        is provided for the message.

        Raises:
            ValueError: If both content and binary are None.
        """
        if not isinstance(self.content, (str, list)):
            raise ValueError(
                f"Unexpected content type: {type(self.content)}. "
                f"Must by a Python string or a list."
            )

    def _iter_content_items(
        self, *, return_text: bool = False, return_images: bool = False
    ) -> Generator[ContentItem, None, None]:
        """Returns a list of content items."""
        if isinstance(self.content, str):
            if return_text:
                yield ContentItem(type=Type.TEXT, content=self.content)
        elif isinstance(self.content, list):
            if return_text and return_images:
                yield from self.content
            else:
                for item in self.content:
                    if (return_text and item.is_text()) or (
                        return_images and item.is_image()
                    ):
                        yield item

    def _iter_all_content_items(self) -> Generator[ContentItem, None, None]:
        return self._iter_content_items(return_text=True, return_images=True)

    def count_content_items(self) -> ContentItemCounts:
        """Counts content items by type."""
        total_items: int = 0
        num_text_items: int = 0
        num_image_items: int = 0
        for item in self._iter_all_content_items():
            total_items += 1
            if item.is_text():
                num_text_items += 1
            elif item.is_image():
                num_image_items += 1

        return ContentItemCounts(
            total_items=total_items,
            text_items=num_text_items,
            image_items=num_image_items,
        )

    @property
    def content_items(self) -> list[ContentItem]:
        """Returns a list of text content items."""
        return [item for item in self._iter_all_content_items()]

    @property
    def image_content_items(self) -> list[ContentItem]:
        """Returns a list of image content items."""
        return [item for item in self._iter_content_items(return_images=True)]

    @property
    def text_content_items(self) -> list[ContentItem]:
        """Returns a list of text content items."""
        return [item for item in self._iter_content_items(return_text=True)]

    def compute_flattened_text_content(self, separator=" ") -> str:
        """Joins contents of all text items."""
        return separator.join(
            [(item.content or "") for item in self.text_content_items]
        )

    def contains_images(self) -> bool:
        """Checks if the message contains at least one image."""
        first_image = next(self._iter_content_items(return_images=True), None)
        return first_image is not None

    def contains_text(self) -> bool:
        """Checks if the message contains at least one text item."""
        first_text = next(self._iter_content_items(return_text=True), None)
        return first_text is not None

    def contains_image_content_items_only(self) -> bool:
        """Checks if the message contains only image items.

        At least one image item is required.
        """
        counts = self.count_content_items()
        return counts.image_items > 0 and counts.image_items == counts.total_items

    def contains_text_content_items_only(self) -> bool:
        """Checks if the message contains only text items.

        At least one text item is required.
        """
        counts = self.count_content_items()
        return counts.text_items > 0 and counts.text_items == counts.total_items

    def contains_single_text_content_item_only(self) -> bool:
        """Checks if the message contains exactly 1 text item, and nothing else.

        These are the most common and simple messages, and may need special handling.
        """
        counts = self.count_content_items()
        return counts.text_items == 1 and counts.text_items == counts.total_items

    def contains_single_image_content_item_only(self) -> bool:
        """Checks if the message contains exactly 1 image item, and nothing else."""
        counts = self.count_content_items()
        return counts.image_items == 1 and counts.image_items == counts.total_items

    @staticmethod
    def from_proto(message_proto: pb2.Message) -> "Message":
        """Converts a Protocol Buffer to a message."""
        if (len(message_proto.parts) == 1) and (
            message_proto.parts[0].type == pb2.ContentPart.TEXT
        ):
            return Message(
                id=(message_proto.id or None),
                role=_convert_proto_role_to_role(message_proto.role),
                content=message_proto.parts[0].content,
            )

        return Message(
            id=(message_proto.id or None),
            role=_convert_proto_role_to_role(message_proto.role),
            content=[ContentItem.from_proto(part) for part in message_proto.parts],
        )

    def to_proto(self) -> pb2.Message:
        """Converts a message to Protocol Buffer format."""
        return pb2.Message(
            id=self.id,
            role=_convert_role_to_proto_role(self.role),
            parts=[item.to_proto() for item in self.content_items],
        )

    def __repr__(self) -> str:
        """Returns a string representation of the message."""
        id_str = ""
        if self.id:
            id_str = f"{self.id} - "
        return f"{id_str}{self.role.upper()}: " + " | ".join(
            [repr(item) for item in self._iter_all_content_items()]
        )


class Conversation(pydantic.BaseModel):
    """Represents a conversation, which is a sequence of messages."""

    conversation_id: Optional[str] = None
    """Optional unique identifier for the conversation.

    This attribute can be used to assign a specific identifier to the conversation,
    which may be useful for tracking or referencing conversations in a larger context.
    """

    messages: list[Message]
    """List of Message objects that make up the conversation."""

    metadata: dict[str, Any] = {}
    """Optional metadata associated with the conversation.

    This attribute allows for storing additional information about the conversation
    in a key-value format. It can be used to include any relevant contextual data.
    """

    def __getitem__(self, idx: int) -> Message:
        """Gets the message at the specified index.

        Args:
            idx (int): The index of the message to retrieve.

        Returns:
            Any: The message at the specified index.
        """
        return self.messages[idx]

    def first_message(self, role: Optional[Role] = None) -> Optional[Message]:
        """Gets the first message in the conversation, optionally filtered by role.

        Args:
            role: The role to filter messages by.
                If None, considers all messages.

        Returns:
            Optional[Message]: The first message matching the criteria,
                or None if no messages are found.
        """
        messages = self.filter_messages(role=role)
        return messages[0] if len(messages) > 0 else None

    def last_message(self, role: Optional[Role] = None) -> Optional[Message]:
        """Gets the last message in the conversation, optionally filtered by role.

        Args:
            role: The role to filter messages by.
                If None, considers all messages.

        Returns:
            Optional[Message]: The last message matching the criteria,
                or None if no messages are found.
        """
        messages = self.filter_messages(role=role)
        return messages[-1] if len(messages) > 0 else None

    def filter_messages(
        self,
        *,
        role: Optional[Role] = None,
        filter_fn: Optional[Callable[[Message], bool]] = None,
    ) -> list[Message]:
        """Gets all messages in the conversation, optionally filtered by role.

        Args:
            role (Optional): The role to filter messages by. If None, no filtering
                by role is applied.
            filter_fn (Optional): A predicate to filter messages by. If the predicate
                returns True for a message, then the message is returned.
                Otherwise, the message is excluded.

        Returns:
            List[Message]: A list of all messages matching the criteria.
        """
        if role is not None:
            messages = [message for message in self.messages if role == message.role]
        else:
            messages = self.messages

        if filter_fn is not None:
            messages = [message for message in messages if filter_fn(message)]

        return messages

    def to_dict(self):
        """Converts the conversation to a dictionary."""
        return self.model_dump(
            mode="json", exclude_unset=True, exclude_defaults=False, exclude_none=True
        )

    def append_id_to_string(self, s: str) -> str:
        """Appends conversation ID to a string.

        Can be useful for log or exception errors messages to allow users
        to identify relevant conversation.
        """
        if not self.conversation_id:
            return s
        suffix = f"Conversation id: {self.conversation_id}."
        return (s.strip() + " " + suffix) if s else suffix

    @classmethod
    def from_dict(cls, data: dict) -> "Conversation":
        """Converts a dictionary to a conversation."""
        return cls.model_validate(data)

    def to_json(self) -> str:
        """Converts the conversation to a JSON string."""
        return self.model_dump_json(
            exclude_unset=True, exclude_defaults=False, exclude_none=True
        )

    @classmethod
    def from_json(cls, data: str) -> "Conversation":
        """Converts a JSON string to a conversation."""
        return cls.model_validate_json(data)

    @staticmethod
    def from_proto(conversation_proto: pb2.Conversation) -> "Conversation":
        """Converts a conversation from Protocol Buffer format."""
        result: Conversation = Conversation(
            conversation_id=(conversation_proto.conversation_id or None),
            messages=[Message.from_proto(m) for m in conversation_proto.messages],
        )
        for key, value in conversation_proto.metadata.items():
            result.metadata[key] = str(value)
        return result

    def to_proto(self) -> pb2.Conversation:
        """Converts a conversation to Protocol Buffer format."""
        result = pb2.Conversation(
            conversation_id=self.conversation_id,
            messages=[m.to_proto() for m in self.messages],
        )
        if self.metadata is not None:
            for key, value in self.metadata.items():
                result.metadata[key] = str(value)
        return result

    def __repr__(self) -> str:
        """Returns a string representation of the conversation."""
        return "\n".join([repr(m) for m in self.messages])


class TemplatedMessage(pydantic.BaseModel):
    """Represents a templated message.

    This class is used to create messages with dynamic content using a template.
    The template can be rendered with variables to produce the final message content.
    """

    template: str
    """The template string used to generate the message content."""

    role: Role
    """The role of the message sender (e.g., USER, ASSISTANT, SYSTEM)."""

    @property
    def content(self) -> str:
        """Renders the content of the message."""
        template = Template(self.template)

        fields = self.model_dump()
        fields.pop("template")  # remove the template from the fields

        return template.render(**fields).strip()

    @property
    def message(self) -> Message:
        """Returns the message in oumi format."""
        content = str(self.content)
        return Message(content=content, role=self.role)



"""Script to copy and clean up documentation files.

This is needed to generate the documentation for the oumi package.

Sphinx requires the documentation files to be in the source tree in order
to build the docs, but many files are better suited to be stored outside
the docs source tree (e.g. notebooks, README files).

Usage examples:

.. code-block:: bash

    # Copy documentation files
    python _manage_doclinks.py copy path/to/doc_links.csv

    # Clean up documentation files
    python _manage_doclinks.py clean path/to/doc_links.csv

    # Adding --absolute-paths will use absolute paths
    # for the source and destination files
    python _manage_doclinks.py copy path/to/doc_links.csv --absolute-paths


The doc_links.csv file should be a CSV file with '|' as the delimiter,
containing two columns: source path and destination path.
"""

import csv
import shutil
from collections.abc import Iterable
from pathlib import Path

import typer

app = typer.Typer()

# Get the root directory of the oumi git repository
OUMI_ROOT: Path = Path(__file__).parent.parent


def read_doc_links(doc_links_config: Path) -> Iterable[tuple[str, str]]:
    """Read the doc links config and yield source and destination paths.

    The doc links file should be a CSV file with '|' as the delimiter.
    Each row should contain two columns: source path and destination path.

    Args:
        doc_links_config (Path): Path to the doc links configuration file.

    Yields:
        tuple: A tuple containing (source_path, destination_path), both as Path objects.
    """
    with doc_links_config.open("r") as f:
        reader = csv.reader(f, delimiter="|")
        for source, destination in reader:
            source = source.strip()
            destination = destination.strip()
            yield source, destination


@app.command()
def copy(
    doc_links_config: Path = typer.Argument(..., help="Path to the doc links config"),
    absolute_paths: bool = typer.Option(
        False, help="Whether to use absolute paths for the source files"
    ),
):
    """Copy documentation files as specified in the doc links config.

    This function reads the doc links configuration file and copies the specified
    files from their source locations to the destination paths. It can handle both
    relative and absolute paths based on the `absolute_paths` parameter.

    Args:
        doc_links_config (Path): Path to the doc links configuration file.
        absolute_paths (bool): If True, use absolute paths for source files.
            If False, paths are relative to OUMI_ROOT.
    """
    if not doc_links_config.exists():
        typer.echo(f"Warning: {doc_links_config} not found. Skipping file copy.")
        return

    for source, destination in read_doc_links(doc_links_config):
        if not absolute_paths:
            source_path = OUMI_ROOT / source
            destination_path = OUMI_ROOT / destination
        else:
            source_path = Path(source)
            destination_path = Path(destination)

        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(source_path, destination_path)
        typer.echo(f"Copied {source_path} to {destination_path}")


@app.command()
def clean(
    doc_links_config: Path = typer.Argument(..., help="Path to the doc links config"),
    absolute_paths: bool = typer.Option(
        False, help="Whether to use absolute paths for the destination files"
    ),
):
    """Delete destination files specified in the doc links config.

    This function reads the doc links configuration file and deletes the specified
    destination files. It can handle both relative and absolute paths based on the
    `absolute_paths` parameter.

    Args:
        doc_links_config (Path): Path to the doc links configuration file.
        absolute_paths (bool): If True, use absolute paths for destination files.
            If False, paths are relative to OUMI_ROOT.
    """
    if not doc_links_config.exists():
        typer.echo(f"Warning: {doc_links_config} not found. Skipping cleanup.")
        return

    for _, destination in read_doc_links(doc_links_config):
        if not absolute_paths:
            destination_path = OUMI_ROOT / destination
        else:
            destination_path = Path(destination)

        if destination_path.exists():
            destination_path.unlink()
            typer.echo(f"Deleted {destination_path}")
        else:
            typer.echo(f"Warning: {destination_path} not found. Skipping deletion.")


if __name__ == "__main__":
    app()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from oumi.core.configs.params.base_params import BaseParams
from oumi.utils.placeholders import get_placeholders, resolve_placeholders


class JudgeResponseFormat(str, Enum):
    """Enumeration of possible response formats for the judge output."""

    JSON = "json"
    """JSON structured response format."""

    XML = "xml"
    """XML-tagged response format."""

    RAW = "raw"
    """Plain text response format."""


class JudgeOutputType(str, Enum):
    """Enumeration of possible output types for the judge's output fields."""

    TEXT = "text"
    """Free-form text judgment."""

    ENUM = "enum"
    """Categorical judgment from predefined options."""

    INT = "int"
    """Integer value judgment."""

    FLOAT = "float"
    """Floating-point value judgment."""

    BOOL = "bool"
    """Boolean judgment (True/False, Yes/No)."""


@dataclass
class JudgeParams(BaseParams):
    """Parameters for the Judge prompt and response format.

    This class holds the parameters for a single-attribute judge,
    including the prompt template and response format.

    Examples:
        Basic boolean judgment:
        >>> judge_params = JudgeParams( # doctest: +SKIP
        ...     prompt_template="Is the following answer helpful? Question: {question},
        ...                      Answer: {answer}. Respond with True or False.",
        ...     response_format=JudgeResponseFormat.XML,
        ...     judgment_type=JudgeOutputType.BOOL,
        ...     include_explanation=False
        ... )

        Categorical judgment with scores:
        >>> judge_params = JudgeParams( # doctest: +SKIP
        ...     prompt_template="Rate the quality of this text: {text}.
        ..                       Respond with 'excellent', 'good', or 'poor'.",
        ...     response_format=JudgeResponseFormat.JSON,
        ...     judgment_type=JudgeOutputType.ENUM,
        ...     judgment_scores={"excellent": 1.0, "good": 0.7, "poor": 0.3},
        ...     include_explanation=True
        ... )
    """

    prompt_template: str
    """Template for the judge prompt with placeholders, such as {question}, {answer}."""

    prompt_template_placeholders: Optional[list[str]] = field(default=None)
    """List of placeholder names in `prompt_template`, to be replaced by the dataset.

    These placeholders correspond to the keys that are expected to be found in every
    example of the input dataset. Their values will replace the placeholders of the
    prompt template, generating a different judge prompt for each example.
    Specifically, if the prompt template contains "{question}" and "{answer}",
    this list (if defined) should be ["question", "answer"].
    """

    system_instruction: Optional[str] = field(default=None)
    """Optional system message to guide judge behavior."""

    template_variables: dict[str, str] = field(default_factory=dict)
    """Variables to be replaced in `prompt_template` and `system_instruction`.

    This dictionary contains variable names and their corresponding values that should
    be replaced in the `prompt_template` and `system_instruction` fields, before the
    dataset-based placeholders are processed. These variables have the following format:
    {variable_name}."""

    response_format: JudgeResponseFormat = field(default=JudgeResponseFormat.XML)
    """The format in which the judge should respond."""

    judgment_type: JudgeOutputType = field(default=JudgeOutputType.BOOL)
    """The type of output that the judgment should be provided with."""

    judgment_scores: Optional[dict[str, float]] = field(default=None)
    """For ENUM judgment_type, the mapping from category names to numeric scores.

    Example:
        {"excellent": 1.0, "good": 0.7, "poor": 0.3}
    """

    include_explanation: bool = field(default=False)
    """Whether the judge should provide an explanation before the judgment."""

    examples: list[dict[str, str]] = field(default_factory=list)
    """Few-shot examples for the judge as a list of field value dictionaries.

    Each dictionary should contain values for all template placeholders and
    expected output fields. Used to provide examples of how the judge should respond.

    Example::

        [
            {
                "question": "What is 2+2?",                      # placeholder value
                "answer": "4",                                   # placeholder value
                "judgment": "Correct",                           # output field value
                "explanation": "It is mathematically correct."   # output field value
            },
            {
                "question": "What is the capital of Mars?",      # placeholder value
                "answer": "New York",                            # placeholder value
                "judgment": "Incorrect",                         # output field value
                "explanation": "Mars does not have capitals."    # output field value
            }
        ]
    """

    def __post_init__(self):
        """Validate the parameters after initialization."""
        self._validate_params()

    def _validate_params(self):
        """Validate the parameters for consistency and completeness.

        Raises:
            ValueError: If parameters are invalid
        """
        # Validate prompt template is not empty
        if not self.prompt_template.strip():
            raise ValueError("prompt_template cannot be empty")

        # Validate judgment scores for ENUM judgment type
        if self.judgment_type == JudgeOutputType.ENUM and not self.judgment_scores:
            raise ValueError("judgment_scores must be provided for ENUM judgment_type")

        # Validate judgment scores are numeric if provided
        if self.judgment_scores:
            if not all(
                isinstance(score, (int, float))
                for score in self.judgment_scores.values()
            ):
                raise ValueError("All judgment_scores values must be numeric")
            if not self.judgment_scores:
                raise ValueError("judgment_scores cannot be empty when provided")

        # Validate prompt_template_placeholders
        if self.prompt_template_placeholders:
            actual_placeholders = self.get_placeholders()
            declared_placeholders = set(self.prompt_template_placeholders)
            if declared_placeholders != actual_placeholders:
                raise ValueError(
                    f"prompt_template_placeholders ({declared_placeholders}) are "
                    "inconsistent with placeholders found in the prompt_template "
                    f"({actual_placeholders})"
                )

    def replace_template_variables(self):
        """Apply template variables to prompt_template and system_instruction."""
        if not self.template_variables:
            return

        self.prompt_template = resolve_placeholders(
            self.prompt_template, self.template_variables, missing_values_allowed=True
        )
        if self.system_instruction:
            self.system_instruction = resolve_placeholders(
                self.system_instruction,
                self.template_variables,
                missing_values_allowed=True,
            )

    def get_placeholders(self) -> set[str]:
        """Get the prompt template placeholders, after template variable replacement."""
        prompt_template = resolve_placeholders(
            self.prompt_template, self.template_variables, missing_values_allowed=True
        )
        return get_placeholders(prompt_template)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.params.synthesis_params import GeneralSynthesisParams
from oumi.core.configs.synthesis_config import SynthesisConfig, SynthesisStrategy


def test_default_synthesis_config():
    """Test default initialization of SynthesisConfig."""
    config = SynthesisConfig()

    assert config.strategy == SynthesisStrategy.GENERAL
    assert isinstance(config.strategy_params, GeneralSynthesisParams)
    assert isinstance(config.inference_config, InferenceConfig)
    assert config.num_samples == 1


def test_custom_synthesis_config():
    """Test custom initialization of SynthesisConfig."""
    custom_params = GeneralSynthesisParams()
    custom_inference = InferenceConfig()

    config = SynthesisConfig(
        strategy=SynthesisStrategy.GENERAL,
        strategy_params=custom_params,
        inference_config=custom_inference,
        num_samples=5,
    )

    assert config.strategy == SynthesisStrategy.GENERAL
    assert config.strategy_params == custom_params
    assert config.inference_config == custom_inference
    assert config.num_samples == 5


def test_invalid_strategy():
    """Test that invalid strategy raises ValueError."""
    config = SynthesisConfig()
    config.strategy = "invalid_strategy"  # type: ignore

    with pytest.raises(ValueError, match="Unsupported synthesis strategy"):
        config.__post_init__()


def test_invalid_input_path():
    """Test that setting input_path raises ValueError."""
    inference_config = InferenceConfig(input_path="some/path")

    with pytest.raises(ValueError, match="Input path is not supported"):
        SynthesisConfig(inference_config=inference_config)


def test_invalid_output_path():
    """Test that setting output_path raises ValueError."""
    inference_config = InferenceConfig(output_path="some/path")

    with pytest.raises(ValueError, match="Output path is not supported"):
        SynthesisConfig(inference_config=inference_config)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.judge_config import (
    JudgeConfig,
)
from oumi.core.configs.params.judge_params import (
    JudgeOutputType,
    JudgeParams,
    JudgeResponseFormat,
)
from oumi.core.configs.params.model_params import ModelParams
from oumi.judges.simple_judge import (
    EXPLANATION_KEY,
    JSON_SUFFIX,
    JSON_SUFFIX_WITH_EXPLANATION,
    JUDGMENT_KEY,
    RAW_SUFFIX_WITH_EXPLANATION,
    XML_SUFFIX,
    XML_SUFFIX_WITH_EXPLANATION,
    SimpleJudge,
)


class TestSimpleJudge:
    """Test cases for the SimpleJudge class."""

    # Test constants
    TEST_PROMPT_TEMPLATE = "Test: {input}"
    TEST_ENUM_SCORES = {"excellent": 1.0, "good": 0.7, "poor": 0.3}

    # Test judge input data
    XML_JUDGE_INPUT = {"question": "What is 2+2?", "answer": "4"}
    JSON_JUDGE_INPUT = {"answer": "Some good answer"}
    RAW_JUDGE_INPUT = {"content": "Some content"}

    # Pre-formatted suffix constants for tests
    BOOL_JUDGMENT_OPTIONS = "Your judgment should be a single word: 'Yes' or 'No'. "
    ENUM_JUDGMENT_OPTIONS = (
        "Your judgment should be one of the following options: "
        "'excellent', 'good', 'poor'. "
    )
    TEXT_JUDGMENT_OPTIONS = (
        "Your judgment should be provided in the form of free text. "
    )

    XML_SUFFIX_FORMATTED_BOOL = XML_SUFFIX.format(
        judgment_key=JUDGMENT_KEY,
        explanation_key=EXPLANATION_KEY,
        judgment_options=BOOL_JUDGMENT_OPTIONS,
    )
    XML_SUFFIX_WITH_EXPLANATION_FORMATTED_BOOL = XML_SUFFIX_WITH_EXPLANATION.format(
        judgment_key=JUDGMENT_KEY,
        explanation_key=EXPLANATION_KEY,
        judgment_options=BOOL_JUDGMENT_OPTIONS,
    )
    JSON_SUFFIX_FORMATTED_ENUM = JSON_SUFFIX.format(
        judgment_key=JUDGMENT_KEY,
        explanation_key=EXPLANATION_KEY,
        judgment_options=ENUM_JUDGMENT_OPTIONS,
    )
    JSON_SUFFIX_WITH_EXPLANATION_FORMATTED_ENUM = JSON_SUFFIX_WITH_EXPLANATION.format(
        judgment_key=JUDGMENT_KEY,
        explanation_key=EXPLANATION_KEY,
        judgment_options=ENUM_JUDGMENT_OPTIONS,
    )
    RAW_SUFFIX_WITH_EXPLANATION_FORMATTED_TEXT = RAW_SUFFIX_WITH_EXPLANATION.format(
        judgment_key=JUDGMENT_KEY,
        explanation_key=EXPLANATION_KEY,
        judgment_options=TEXT_JUDGMENT_OPTIONS,
    )

    @pytest.fixture
    def xml_config_no_explanation(self):
        return JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=(
                    "Is this helpful? Question: {question}, Answer: {answer}"
                ),
                response_format=JudgeResponseFormat.XML,
                judgment_type=JudgeOutputType.BOOL,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

    @pytest.fixture
    def xml_config_with_explanation(self):
        return JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=(
                    "Is this helpful? Question: {question}, Answer: {answer}"
                ),
                response_format=JudgeResponseFormat.XML,
                judgment_type=JudgeOutputType.BOOL,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

    @pytest.fixture
    def json_config_no_explanation(self):
        return JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Rate this answer: {answer}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores=self.TEST_ENUM_SCORES,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

    @pytest.fixture
    def json_config_with_explanation(self):
        return JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Rate this answer: {answer}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores=self.TEST_ENUM_SCORES,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

    @pytest.fixture
    def mock_inference_engine(self):
        return Mock()

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_init_with_engine_no_explanation(
        self, mock_create_engine, xml_config_no_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=xml_config_no_explanation)

        # Verify the config was properly processed
        assert judge._judge_params == xml_config_no_explanation.judge_params
        assert judge._inference_config == xml_config_no_explanation.inference_config
        assert judge.inference_engine == mock_engine
        assert (
            judge.prompt_template
            == xml_config_no_explanation.judge_params.prompt_template
        )
        assert (
            judge.response_format
            == xml_config_no_explanation.judge_params.response_format
        )

        # Should have one output field for judgment
        assert len(judge.output_fields) == 1
        assert judge.output_fields[0].field_key == JUDGMENT_KEY
        assert judge.output_fields[0].field_type == JudgeOutputType.BOOL
        assert judge.output_fields[0].field_scores is None

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_init_with_engine_explanation(
        self, mock_create_engine, json_config_with_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=json_config_with_explanation)

        # Should have two output fields: explanation and judgment
        assert len(judge.output_fields) == 2
        assert judge.output_fields[0].field_key == EXPLANATION_KEY
        assert judge.output_fields[0].field_type == JudgeOutputType.TEXT
        assert judge.output_fields[0].field_scores is None
        assert judge.output_fields[1].field_key == JUDGMENT_KEY
        assert judge.output_fields[1].field_type == JudgeOutputType.ENUM
        assert judge.output_fields[1].field_scores == {
            "excellent": 1.0,
            "good": 0.7,
            "poor": 0.3,
        }

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_init_creates_inference_engine(
        self, mock_create_engine, xml_config_no_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=xml_config_no_explanation)

        assert judge.inference_engine == mock_engine
        mock_create_engine.assert_called_once_with(
            xml_config_no_explanation.inference_config
        )

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_xml_no_explanation(
        self, mock_create_engine, xml_config_no_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=xml_config_no_explanation)

        prompt = judge._build_judgment_prompt(self.XML_JUDGE_INPUT)

        expected = (
            f"Is this helpful? Question: What is 2+2?, Answer: 4"
            f"{self.XML_SUFFIX_FORMATTED_BOOL}"
        )
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_xml_with_explanation(
        self, mock_create_engine, xml_config_with_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=xml_config_with_explanation)

        prompt = judge._build_judgment_prompt(self.XML_JUDGE_INPUT)

        expected = (
            f"Is this helpful? Question: What is 2+2?, Answer: 4"
            f"{self.XML_SUFFIX_WITH_EXPLANATION_FORMATTED_BOOL}"
        )
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_json_no_explanation(
        self, mock_create_engine, json_config_no_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=json_config_no_explanation)

        prompt = judge._build_judgment_prompt(self.JSON_JUDGE_INPUT)

        expected = (
            f"Rate this answer: Some good answer{self.JSON_SUFFIX_FORMATTED_ENUM}"
        )
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_json_with_explanation(
        self, mock_create_engine, json_config_with_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=json_config_with_explanation)

        prompt = judge._build_judgment_prompt(self.JSON_JUDGE_INPUT)

        expected = (
            f"Rate this answer: Some good answer"
            f"{self.JSON_SUFFIX_WITH_EXPLANATION_FORMATTED_ENUM}"
        )
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_raw_no_explanation(self, mock_create_engine):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        config = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Evaluate: {content}",
                response_format=JudgeResponseFormat.RAW,
                judgment_type=JudgeOutputType.TEXT,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge = SimpleJudge(judge_config=config)

        prompt = judge._build_judgment_prompt(self.RAW_JUDGE_INPUT)

        expected = "Evaluate: Some content"  # No suffix for RAW without explanation
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_build_prompt_raw_with_explanation(self, mock_create_engine):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        config = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Evaluate: {content}",
                response_format=JudgeResponseFormat.RAW,
                judgment_type=JudgeOutputType.TEXT,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge = SimpleJudge(judge_config=config)

        prompt = judge._build_judgment_prompt(self.RAW_JUDGE_INPUT)

        expected = (
            f"Evaluate: Some content{self.RAW_SUFFIX_WITH_EXPLANATION_FORMATTED_TEXT}"
        )
        assert prompt == expected

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_get_format_suffix_xml(self, mock_create_engine):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        config_no_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.XML,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_no_exp = SimpleJudge(judge_config=config_no_exp)
        assert judge_no_exp._get_format_suffix() == self.XML_SUFFIX_FORMATTED_BOOL

        config_with_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.XML,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_with_exp = SimpleJudge(judge_config=config_with_exp)
        assert (
            judge_with_exp._get_format_suffix()
            == self.XML_SUFFIX_WITH_EXPLANATION_FORMATTED_BOOL
        )

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_get_format_suffix_json(self, mock_create_engine):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        config_no_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores=self.TEST_ENUM_SCORES,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_no_exp = SimpleJudge(judge_config=config_no_exp)
        assert judge_no_exp._get_format_suffix() == self.JSON_SUFFIX_FORMATTED_ENUM

        config_with_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores=self.TEST_ENUM_SCORES,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_with_exp = SimpleJudge(judge_config=config_with_exp)
        assert (
            judge_with_exp._get_format_suffix()
            == self.JSON_SUFFIX_WITH_EXPLANATION_FORMATTED_ENUM
        )

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_get_format_suffix_raw(self, mock_create_engine):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        config_no_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.RAW,
                judgment_type=JudgeOutputType.TEXT,
                include_explanation=False,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_no_exp = SimpleJudge(judge_config=config_no_exp)
        assert judge_no_exp._get_format_suffix() == ""

        config_with_exp = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template=self.TEST_PROMPT_TEMPLATE,
                response_format=JudgeResponseFormat.RAW,
                judgment_type=JudgeOutputType.TEXT,
                include_explanation=True,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )
        judge_with_exp = SimpleJudge(judge_config=config_with_exp)
        assert (
            judge_with_exp._get_format_suffix()
            == self.RAW_SUFFIX_WITH_EXPLANATION_FORMATTED_TEXT
        )

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_create_judgment_output_field(
        self, mock_create_engine, xml_config_no_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=xml_config_no_explanation)

        field = judge._create_judgment_output_field(
            xml_config_no_explanation.judge_params
        )

        assert field.field_key == JUDGMENT_KEY
        assert field.field_type == JudgeOutputType.BOOL
        assert field.field_scores is None

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_create_judgment_output_field_with_scores(
        self, mock_create_engine, json_config_with_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=json_config_with_explanation)

        field = judge._create_judgment_output_field(
            json_config_with_explanation.judge_params
        )

        assert field.field_key == JUDGMENT_KEY
        assert field.field_type == JudgeOutputType.ENUM
        assert field.field_scores == {"excellent": 1.0, "good": 0.7, "poor": 0.3}

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_create_explanation_output_field(
        self, mock_create_engine, json_config_with_explanation
    ):
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        judge = SimpleJudge(judge_config=json_config_with_explanation)

        field = judge._create_explanation_output_field()

        assert field.field_key == EXPLANATION_KEY
        assert field.field_type == JudgeOutputType.TEXT
        assert field.field_scores is None

    @patch("oumi.builders.inference_engines.build_inference_engine")
    def test_create_inference_engine(
        self, mock_build_engine, xml_config_no_explanation
    ):
        mock_engine = Mock()
        mock_build_engine.return_value = mock_engine

        _ = SimpleJudge(judge_config=xml_config_no_explanation)

        # Should be called during init with inference_config
        mock_build_engine.assert_called_once_with(
            engine_type=xml_config_no_explanation.inference_config.engine,
            model_params=xml_config_no_explanation.inference_config.model,
            remote_params=xml_config_no_explanation.inference_config.remote_params,
            generation_params=xml_config_no_explanation.inference_config.generation,
        )

    def test_enum_judgment_type_requires_scores(self):
        """Test that ENUM judgment type requires judgment_scores to be provided."""
        with pytest.raises(
            ValueError, match="judgment_scores must be provided for ENUM judgment_type"
        ):
            JudgeParams(
                prompt_template="Rate this: {text}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores=None,
            )

    def test_enum_judgment_type_with_empty_scores(self):
        """Test that ENUM judgment type with empty scores fails validation."""
        with pytest.raises(
            ValueError, match="judgment_scores must be provided for ENUM judgment_type"
        ):
            JudgeParams(
                prompt_template="Rate this: {text}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.ENUM,
                judgment_scores={},
            )

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_format_suffix_behavior_with_and_without_system_instruction(
        self, mock_create_engine
    ):
        """Test that format suffix goes to system instruction or judgment prompt."""
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        # Test case 1: With system instruction - format suffix in system instruction
        config_with_system = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Rate: {text}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.BOOL,
                include_explanation=False,
                system_instruction="You are a judge.",
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

        judge = SimpleJudge(judge_config=config_with_system)

        # System instruction should have format suffix
        assert judge.system_instruction is not None
        assert "JSON format only" in judge.system_instruction
        assert judge.system_instruction.startswith("You are a judge.")

        # Judgment prompt should NOT have format suffix
        prompt_with_system = judge._build_judgment_prompt({"text": "test"})
        assert prompt_with_system == "Rate: test"  # No format suffix here

        # Test case 2: Without system instruction - format suffix in judgment prompt
        config_without_system = JudgeConfig(
            judge_params=JudgeParams(
                prompt_template="Rate: {text}",
                response_format=JudgeResponseFormat.JSON,
                judgment_type=JudgeOutputType.BOOL,
                include_explanation=False,
                system_instruction=None,
            ),
            inference_config=InferenceConfig(
                engine=InferenceEngineType.NATIVE,
                model=ModelParams(model_name="test-model"),
            ),
        )

        judge = SimpleJudge(judge_config=config_without_system)

        # System instruction should be None
        assert judge.system_instruction is None

        # Judgment prompt should have format suffix
        prompt_without_system = judge._build_judgment_prompt({"text": "test"})
        assert "JSON format only" in prompt_without_system
        assert prompt_without_system.startswith("Rate: test")

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_init_with_judge_config_path(self, mock_create_engine):
        """Test initialization with a built-in judge string."""
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        yaml_judge_config = """
judge_params:
  prompt_template: "Is the following Q&A correct? Question: {question} Answer: {answer}"
  response_format: JSON
  judgment_type: BOOL
  include_explanation: False
inference_config:
  model:
    model_name: "test-model"
  engine: NATIVE
"""
        with tempfile.TemporaryDirectory() as temp_dir:
            judge_config_path = str(Path(temp_dir) / "judge_config.yaml")
            Path(judge_config_path).write_text(yaml_judge_config)

            judge = SimpleJudge(judge_config=judge_config_path)

            # Should resolve to the built-in config
            assert judge._judge_params is not None
            assert judge._judge_params.judgment_type == JudgeOutputType.BOOL
            assert judge._judge_params.include_explanation is False
            assert judge._judge_params.response_format == JudgeResponseFormat.JSON

    @patch("oumi.judges.simple_judge.SimpleJudge._create_inference_engine")
    def test_init_with_unknown_judge_config_path(self, mock_create_engine):
        """Test initialization with an unknown built-in judge string raises error."""
        mock_engine = Mock()
        mock_create_engine.return_value = mock_engine

        with pytest.raises(
            ValueError, match="Could not resolve JudgeConfig from path: unknown_judge"
        ):
            SimpleJudge(judge_config="unknown_judge")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GRPO reward functions module."""

from oumi.datasets.grpo.rewards.completion_length_rewards import (
    compute_sharp_target_token_length_reward,
    compute_soft_target_token_length_reward,
)
from oumi.datasets.grpo.rewards.count_letters_rewards import compute_letter_count_reward
from oumi.datasets.grpo.rewards.countdown_rewards import countdown_reward
from oumi.datasets.grpo.rewards.gsm8k_reward import gsm8k_reward

__all__ = [
    "compute_letter_count_reward",
    "compute_soft_target_token_length_reward",
    "compute_sharp_target_token_length_reward",
    "countdown_reward",
    "gsm8k_reward",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union, cast

from PIL import Image
from typing_extensions import override

from oumi.builders.processors import build_processor
from oumi.core.configs.internal.internal_model_config import (
    InternalFeatureFirstDimAction,
    InternalModelConfig,
)
from oumi.core.configs.internal.supported_models import (
    find_internal_model_config_using_model_name,
    get_default_vlm_model_config,
)
from oumi.core.datasets.base_dpo_dataset import BaseDpoDataset
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types.conversation import ContentItem, Role, Type
from oumi.utils.conversation_utils import load_pil_image_from_content_item

_PROMPT_KEY = "prompt"
_CHOSEN_KEY = "chosen"
_REJECTED_KEY = "rejected"
_IMAGES_KEY = "images"


class VisionLanguageDpoDataset(BaseDpoDataset):
    """Dataset for vision-language DPO (Direct Preference Optimization) models.

    This class extends BaseDpoDataset to provide functionality specific to
    vision-language preference optimization tasks. It handles the processing of
    both image and text data for preference learning.

    The dataset expects data in the formats::

        {
            "prompt": "What's in this image?",
            "images": ["path/to/image.jpg", ...],  # Optional image paths/URLs
            "chosen": [{"role": "assistant", "content": "I see a cat"}],
            "rejected": [{"role": "assistant", "content": "I see a dog"}]
        }

        OR

        {
            "prompt": "What's in this image?",aths/URLs
            "images": ["path/to/image.jpg", ...],
            "chosen": "preferred response",
            "rejected": "rejected response"
        }
    """

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[str] = None,
        split: Optional[str] = None,
        tokenizer: Optional[BaseTokenizer] = None,
        return_tensors: bool = False,
        processor: Optional[Any] = None,
        processor_name: Optional[str] = None,
        trust_remote_code: bool = False,
        processor_kwargs: Optional[dict[str, Any]] = None,
        max_size: Optional[int] = None,
        prompt_key: str = _PROMPT_KEY,
        chosen_key: str = _CHOSEN_KEY,
        rejected_key: str = _REJECTED_KEY,
        images_key: str = _IMAGES_KEY,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the VisionLanguageDpoDataset class.

        The dataset will return dictionaries containing formatted preference data
        ready for DPO training with chat templates applied.

        Args:
            processor: The vision-language processor for applying chat templates
                and processing images.
            tokenizer: The tokenizer for encoding text data.
            return_tensors: Whether to return tensors instead of strings.
            dataset_name: The name of the dataset.
            dataset_path: The path to the dataset.
            split: The split of the dataset.
            processor_name: The name of the processor to use.
            trust_remote_code: Whether to trust remote code.
            processor_kwargs: Additional keyword arguments to pass to the processor.
            max_size: The maximum size of the longest edge of the image in pixels.
            prompt_key: The key for the prompt.
            chosen_key: The key for the chosen.
            rejected_key: The key for the rejected.
            images_key: The key for the images.
            **kwargs: Additional keyword arguments to pass to the base class.
        """
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=dataset_path,
            split=split,
            tokenizer=tokenizer,
            return_tensors=return_tensors,
            **kwargs,
        )

        # Build the processor.
        if processor is not None:
            self._processor = processor
        elif processor_name is not None and self._tokenizer is not None:
            self._processor = build_processor(
                processor_name,
                self._tokenizer,
                trust_remote_code=True,
                processor_kwargs=processor_kwargs,
            )
        else:
            raise ValueError("Processor is not set.")

        self._internal_model_config: InternalModelConfig = (
            find_internal_model_config_using_model_name(
                self._processor.processor_name, trust_remote_code=True
            )
            or get_default_vlm_model_config()
        )
        self._max_size = max_size
        self._input_prompt_key = prompt_key
        self._input_chosen_key = chosen_key
        self._input_rejected_key = rejected_key
        self._input_images_key = images_key

    @override
    def transform_preference(self, sample: dict) -> dict:
        """Transform a DPO sample to the format expected by DPO trainer.

        Transforms a raw DPO example into three Oumi Conversation objects.

        Args:
            sample (dict): A dictionary representing a single DPO preference example.

        Returns:
            Dict with prompt, chosen, and rejected conversations or features
        """
        # First, convert the prompt, chosen, and rejected to conversation dictionaries.
        prompt_chat = self._to_messages_list(sample[self._input_prompt_key], Role.USER)
        chosen_chat = self._to_messages_list(
            sample[self._input_chosen_key], Role.ASSISTANT
        )
        rejected_chat = self._to_messages_list(
            sample[self._input_rejected_key], Role.ASSISTANT
        )
        images = sample[self._input_images_key] or []
        images = [images] if isinstance(images, dict) else images

        # Load and resize the images.
        if images is not None:
            images = [self._resize_image(self._load_image(image)) for image in images]

        # Add the image turns to the prompt if not already present.
        if all(isinstance(turn["content"], str) for turn in prompt_chat):
            for image in images:
                prompt_chat.append(
                    {
                        "role": "user",
                        "content": [{"type": "image_bytes", "content": image}],
                    }
                )

        return {
            _PROMPT_KEY: prompt_chat,
            _CHOSEN_KEY: chosen_chat,
            _REJECTED_KEY: rejected_chat,
            _IMAGES_KEY: images,
        }

    def _load_image(self, image_path: Union[str, ContentItem, dict]) -> Image.Image:
        """Load images from the given paths."""
        if isinstance(image_path, str):
            content_type = (
                Type.IMAGE_URL if image_path.startswith("http") else Type.IMAGE_PATH
            )
            image = ContentItem(type=content_type, content=image_path)
        elif isinstance(image_path, dict):
            image = ContentItem(type=Type.IMAGE_BINARY, binary=image_path["bytes"])
        else:
            image = image_path

        return load_pil_image_from_content_item(image)

    def _resize_image(self, image: Image.Image) -> Image.Image:
        if self._processor is None:
            return image

        if self._max_size is not None:
            image.thumbnail((self._max_size, self._max_size))

        return image

    def _drop_first_dim_if_needed(self, feature_name: str, value: Any) -> Any:
        """Drop the first dimension of the features."""
        feature_spec = self._internal_model_config.model_input_features.get(
            feature_name,
        )

        if feature_spec is None:
            action = InternalFeatureFirstDimAction.DROP_IF_DUMMY
        else:
            action = feature_spec.first_dim_action

        if action == InternalFeatureFirstDimAction.DROP_ALWAYS:
            return value[0]
        elif action == InternalFeatureFirstDimAction.DROP_IF_DUMMY:
            if len(value) == 1:
                return value[0]
            else:
                return value
        return value

    @override
    def _process_sample(
        self,
        features,
    ):
        """Process a row of the dataset."""
        if self._tokenizer is None or self._processor is None:
            raise ValueError(
                "Tokenizer and processor are required to process a sample."
            )

        # Apply the chat template to the prompt.
        prompt = self._tokenizer.apply_chat_template(features["prompt"], tokenize=False)
        prompt = cast(str, prompt)

        # Apply the chat template to the chosen and rejected turns.
        # To get only the completion part, we tokenizer the prompt + chosen/rejected
        # and then remove the prompt prefix.
        prompt_chosen = self._tokenizer.apply_chat_template(
            features["prompt"] + features["chosen"], tokenize=False
        )
        prompt_chosen = cast(str, prompt_chosen)
        chosen = prompt_chosen[len(prompt) :]

        prompt_rejected = self._tokenizer.apply_chat_template(
            features["prompt"] + features["rejected"], tokenize=False
        )
        prompt_rejected = cast(str, prompt_rejected)
        rejected = prompt_rejected[len(prompt) :]

        # Tokenize the prompt, chosen, and rejected turns.
        processed_features = self._processor(images=features["images"], text=[prompt])

        prompt_input_ids = self._drop_first_dim_if_needed(
            "input_ids", processed_features["input_ids"]
        )

        chosen_input_ids = self._tokenizer(chosen, add_special_tokens=False)[
            "input_ids"
        ]
        chosen_input_ids = cast(list[int], chosen_input_ids)
        rejected_input_ids = self._tokenizer(rejected, add_special_tokens=False)[
            "input_ids"
        ]
        rejected_input_ids = cast(list[int], rejected_input_ids)
        chosen_input_ids = chosen_input_ids + [self._tokenizer.eos_token_id]
        rejected_input_ids = rejected_input_ids + [self._tokenizer.eos_token_id]

        output = {
            "prompt_input_ids": prompt_input_ids,
            "chosen_input_ids": chosen_input_ids,
            "rejected_input_ids": rejected_input_ids,
        }

        # Drop the first dimension of the features if needed.
        if "pixel_values" in processed_features:
            output["pixel_values"] = self._drop_first_dim_if_needed(
                "pixel_values", processed_features["pixel_values"]
            )
        if "pixel_attention_mask" in processed_features:
            output["pixel_attention_mask"] = self._drop_first_dim_if_needed(
                "pixel_attention_mask", processed_features["pixel_attention_mask"]
            )
        if "image_sizes" in processed_features:
            output["image_sizes"] = self._drop_first_dim_if_needed(
                "image_sizes", processed_features["image_sizes"]
            )

        return output

    @override
    def transform(self, sample: dict) -> dict:
        """Transform the sample to the Oumi format."""
        return self._process_sample(self.transform_preference(sample))



import copy
from unittest.mock import MagicMock, patch

import pytest
from lm_eval.api.task import ConfigurableTask

from oumi.core.configs import (
    EvaluationConfig,
    GenerationParams,
    InferenceEngineType,
    LMHarnessTaskParams,
    ModelParams,
    RemoteParams,
)
from oumi.core.evaluation.backends.lm_harness import _generate_lm_harness_model_args
from oumi.core.evaluation.backends.lm_harness import evaluate as evaluate_lm_harness


@pytest.fixture
def mock_patches_for_evaluate():
    with (
        patch(
            "oumi.core.evaluation.backends.lm_harness.lm_harness_log_utils"
        ) as mock_lm_harness_log_utils,
        patch(
            "oumi.core.evaluation.backends.lm_harness.WandbLogger"
        ) as mock_WandbLogger,
        patch(
            "oumi.core.evaluation.backends.lm_harness.is_world_process_zero"
        ) as mock_is_world_process_zero,
        patch(
            "oumi.core.evaluation.backends.lm_harness.lm_harness_evaluate"
        ) as mock_lm_harness_evaluate,
        patch(
            "oumi.core.evaluation.backends.lm_harness.lm_harness_get_model_class"
        ) as mock_lm_harness_get_model_class,
        patch(
            "oumi.core.evaluation.backends.lm_harness._generate_lm_harness_model_args"
        ) as mock_generate_lm_harness_model_args,
        patch(
            "oumi.core.evaluation.backends.lm_harness._get_task_dict"
        ) as mock_get_task_dict,
        patch(
            "oumi.core.evaluation.backends.lm_harness.is_image_text_llm_using_model_name"
        ) as mock_is_image_text_llm,
        patch(
            "oumi.core.evaluation.backends.lm_harness._set_random_seeds"
        ) as mock_set_random_seeds,
        patch("torch.cuda.is_available") as mock_cuda_is_available,
    ):
        yield {
            "mock_cuda_is_available": mock_cuda_is_available,
            "mock_set_random_seeds": mock_set_random_seeds,
            "mock_is_image_text_llm": mock_is_image_text_llm,
            "mock_get_task_dict": mock_get_task_dict,
            "mock_generate_lm_harness_model_args": mock_generate_lm_harness_model_args,
            "mock_lm_harness_get_model_class": mock_lm_harness_get_model_class,
            "mock_lm_harness_evaluate": mock_lm_harness_evaluate,
            "mock_is_world_process_zero": mock_is_world_process_zero,
            "mock_WandbLogger": mock_WandbLogger,
            "mock_lm_harness_log_utils": mock_lm_harness_log_utils,
        }


@pytest.mark.parametrize(
    "lm_harness_model, is_multimodal, device, model_params, generation_params, "
    "inference_engine_type, inference_remote_params, expected_model_args",
    [
        (
            "hf",
            False,
            "mps",
            ModelParams(model_name="text_model"),
            GenerationParams(batch_size=None),
            InferenceEngineType.NATIVE,
            None,
            {
                "trust_remote_code": False,
                "pretrained": "text_model",
                "dtype": "auto",
                "max_length": None,
                "batch_size": 1,
                "max_batch_size": None,
                "device": "mps",
                "parallelize": False,
                "device_map": "auto",
            },
        ),
        (
            "hf-multimodal",
            True,
            "cuda:0",
            ModelParams(model_name="vision_model", model_max_length=128),
            GenerationParams(),
            InferenceEngineType.NATIVE,
            None,
            {
                "trust_remote_code": False,
                "pretrained": "vision_model",
                "dtype": "auto",
                "max_length": 128,
                "batch_size": 1,
                "max_batch_size": None,
                "device": "cuda:0",
                "parallelize": False,
                "device_map": "auto",
                "max_images": 1,
                "interleave": True,
                "convert_img_format": True,
                "image_string": "my_image_token",
                "image_token_id": 1111,
            },
        ),
        (
            "vllm",
            False,
            "cuda:0",
            ModelParams(model_name="text_model", model_max_length=128),
            GenerationParams(batch_size=1),
            InferenceEngineType.VLLM,
            None,
            {
                "trust_remote_code": False,
                "pretrained": "text_model",
                "dtype": "auto",
                "max_length": 128,
                "batch_size": 1,
                "max_batch_size": None,
                "device": "cuda:0",
            },
        ),
        (
            "vllm-vlm",
            True,
            "cuda:0",
            ModelParams(
                model_name="vision_model", model_max_length=128, trust_remote_code=True
            ),
            GenerationParams(batch_size=8),
            InferenceEngineType.VLLM,
            None,
            {
                "trust_remote_code": True,
                "pretrained": "vision_model",
                "dtype": "auto",
                "max_length": 128,
                "batch_size": 8,
                "max_batch_size": None,
                "device": "cuda:0",
                "max_images": 1,
                "interleave": True,
            },
        ),
        (
            "local-completions",
            False,
            "cpu",
            ModelParams(model_name="some_model"),
            GenerationParams(),
            InferenceEngineType.REMOTE,
            RemoteParams(
                api_url="http://localhost:6864/v1/completions",
                num_workers=16,
                max_retries=3,
                connection_timeout=120,
            ),
            {
                "trust_remote_code": False,
                "pretrained": "some_model",
                "dtype": "auto",
                "max_length": None,
                "batch_size": 1,
                "max_batch_size": None,
                "device": "cpu",
                "base_url": "http://localhost:6864/v1/completions",
                "num_concurrent": 16,
                "max_retries": 3,
                "timeout": 120,
            },
        ),
    ],
    ids=[
        "model_args_hf_native",
        "model_args_hf-multimodal_native",
        "model_args_vllm",
        "model_args_vllm-vlm",
        "model_args_local-completions",
    ],
)
@patch("oumi.core.evaluation.backends.lm_harness.build_tokenizer")
@patch("oumi.core.evaluation.backends.lm_harness.build_processor")
def test_generate_lm_harness_model_args(
    mock_build_processor,
    mock_build_tokenizer,
    lm_harness_model,
    is_multimodal,
    device,
    model_params,
    generation_params,
    inference_engine_type,
    inference_remote_params,
    expected_model_args,
):
    mock_build_tokenizer.return_value = MagicMock()
    mock_build_processor.return_value = MagicMock(
        image_token="my_image_token", image_token_id=1111
    )

    model_args = _generate_lm_harness_model_args(
        lm_harness_model,
        is_multimodal,
        device,
        model_params,
        generation_params,
        inference_engine_type,
        inference_remote_params,
    )

    if is_multimodal and inference_engine_type == InferenceEngineType.NATIVE:
        mock_build_tokenizer.assert_called_once_with(model_params)
        mock_build_processor.assert_called_once_with(
            model_params.model_name,
            mock_build_tokenizer.return_value,
            trust_remote_code=model_params.trust_remote_code,
            processor_kwargs={},
        )
    else:
        mock_build_tokenizer.assert_not_called()
        mock_build_processor.assert_not_called()

    assert model_args == expected_model_args


def test_evaluate(mock_patches_for_evaluate):
    # Access the relevant mocks through the fixture.
    mock_cuda_is_available = mock_patches_for_evaluate["mock_cuda_is_available"]
    mock_set_random_seeds = mock_patches_for_evaluate["mock_set_random_seeds"]
    mock_is_image_text_llm = mock_patches_for_evaluate["mock_is_image_text_llm"]
    mock_get_task_dict = mock_patches_for_evaluate["mock_get_task_dict"]
    mock_generate_lm_harness_model_args = mock_patches_for_evaluate[
        "mock_generate_lm_harness_model_args"
    ]
    mock_lm_harness_get_model_class = mock_patches_for_evaluate[
        "mock_lm_harness_get_model_class"
    ]
    mock_lm_harness_evaluate = mock_patches_for_evaluate["mock_lm_harness_evaluate"]
    mock_is_world_process_zero = mock_patches_for_evaluate["mock_is_world_process_zero"]

    # Set the inputs of evaluate() function.
    task_params = LMHarnessTaskParams(
        evaluation_backend="lm_harness",
        task_name="mmlu",
        num_samples=222,
    )
    evaluation_config = EvaluationConfig(
        tasks=[task_params],
        model=ModelParams(model_name="gpt2"),
        generation=GenerationParams(),
        inference_engine=InferenceEngineType.NATIVE,
        inference_remote_params=None,
        run_name="run_name",
        enable_wandb=False,
        output_dir="test_output",
    )
    evaluation_config_without_tasks = copy.deepcopy(evaluation_config)
    evaluation_config_without_tasks.tasks = []
    random_seed = 123
    numpy_random_seed = 1234
    torch_random_seed = 12345

    # Mock the outputs of functions that evaluate() calls.
    mock_task_dict = {"mmlu": MagicMock(spec=ConfigurableTask)}
    mock_lm_harness_model_args = {"pretrained": "gpt2"}
    mock_results = {
        "results": {"mmlu": {"acc": 0.77}},
        "configs": {"my_config": "some_config"},
    }

    # Mock functions that evaluate() calls.
    mock_cuda_is_available.return_value = True
    mock_is_image_text_llm.return_value = False
    mock_get_task_dict.return_value = mock_task_dict
    mock_generate_lm_harness_model_args.return_value = mock_lm_harness_model_args
    mock_lm_harness_get_model_class.return_value = MagicMock()
    mock_lm_harness_evaluate.return_value = copy.deepcopy(mock_results)
    mock_is_world_process_zero.return_value = True

    _ = evaluate_lm_harness(
        task_params=task_params,
        config=evaluation_config,
        random_seed=random_seed,
        numpy_random_seed=numpy_random_seed,
        torch_random_seed=torch_random_seed,
    )

    # Assertions
    mock_set_random_seeds.assert_called_once_with(
        random_seed=random_seed,
        numpy_random_seed=numpy_random_seed,
        torch_random_seed=torch_random_seed,
    )
    mock_is_image_text_llm.assert_called_once_with(
        model_name=evaluation_config.model.model_name,
        trust_remote_code=evaluation_config.model.trust_remote_code,
    )
    mock_get_task_dict.assert_called_once_with(task_params)
    mock_generate_lm_harness_model_args.assert_called_once_with(
        lm_harness_model="hf",
        is_multimodal=False,
        device="cuda:0",
        model_params=evaluation_config.model,
        generation_params=evaluation_config.generation,
        inference_engine_type=evaluation_config.inference_engine,
        inference_remote_params=evaluation_config.inference_remote_params,
    )
    mock_lm_harness_get_model_class.assert_called_once_with("hf")

    mock_lm_harness_evaluate.assert_called_once()
    args, kwargs = mock_lm_harness_evaluate.call_args
    assert args[1] == mock_task_dict  # task_dict is now the second positional argument
    assert kwargs["limit"] == 222
    assert not kwargs["apply_chat_template"]


def test_evaluate_failure_vLLM_without_CUDA(mock_patches_for_evaluate):
    # Access the relevant mocks through the fixture.
    mock_cuda_is_available = mock_patches_for_evaluate["mock_cuda_is_available"]
    mock_is_image_text_llm = mock_patches_for_evaluate["mock_is_image_text_llm"]
    mock_get_task_dict = mock_patches_for_evaluate["mock_get_task_dict"]
    mock_generate_lm_harness_model_args = mock_patches_for_evaluate[
        "mock_generate_lm_harness_model_args"
    ]
    mock_lm_harness_get_model_class = mock_patches_for_evaluate[
        "mock_lm_harness_get_model_class"
    ]
    mock_is_world_process_zero = mock_patches_for_evaluate["mock_is_world_process_zero"]

    # This combination should throw (we cannot use VLLM without CUDA).
    inference_engine_type = InferenceEngineType.VLLM
    mock_cuda_is_available.return_value = False

    # Mock functions that evaluate() calls.
    mock_is_image_text_llm.return_value = False
    mock_is_world_process_zero.return_value = True
    mock_get_task_dict.return_value = MagicMock()
    mock_generate_lm_harness_model_args.return_value = MagicMock()
    mock_lm_harness_get_model_class.return_value = MagicMock()

    with pytest.raises(
        ValueError, match="The `VLLM` inference_engine requires a CUDA-enabled GPU."
    ):
        evaluate_lm_harness(
            task_params=LMHarnessTaskParams(
                evaluation_backend="lm_harness",
                task_name="mmlu",
            ),
            config=EvaluationConfig(
                tasks=[],
                inference_engine=inference_engine_type,
            ),
        )


@pytest.mark.parametrize(
    "unsupported_inference_engine_type",
    [
        InferenceEngineType.REMOTE_VLLM,
        InferenceEngineType.SGLANG,
        InferenceEngineType.LLAMACPP,
        InferenceEngineType.ANTHROPIC,
        InferenceEngineType.GOOGLE_VERTEX,
        InferenceEngineType.GOOGLE_GEMINI,
        InferenceEngineType.DEEPSEEK,
        InferenceEngineType.PARASAIL,
        InferenceEngineType.TOGETHER,
        InferenceEngineType.OPENAI,
        InferenceEngineType.SAMBANOVA,
    ],
    ids=[
        "non_supported_engine_remote_vllm",
        "non_supported_engine_sglang",
        "non_supported_engine_llamacpp",
        "non_supported_engine_anthropic",
        "non_supported_engine_google_vertex",
        "non_supported_engine_google_gemini",
        "non_supported_engine_deepseek",
        "non_supported_engine_parasail",
        "non_supported_engine_together",
        "non_supported_engine_openai",
        "non_supported_engine_sambanova",
    ],
)
def test_evaluate_failure_non_supported_engine(
    mock_patches_for_evaluate, unsupported_inference_engine_type
):
    # Access the relevant mocks through the fixture.
    mock_cuda_is_available = mock_patches_for_evaluate["mock_cuda_is_available"]
    mock_is_image_text_llm = mock_patches_for_evaluate["mock_is_image_text_llm"]
    mock_get_task_dict = mock_patches_for_evaluate["mock_get_task_dict"]
    mock_generate_lm_harness_model_args = mock_patches_for_evaluate[
        "mock_generate_lm_harness_model_args"
    ]
    mock_lm_harness_get_model_class = mock_patches_for_evaluate[
        "mock_lm_harness_get_model_class"
    ]
    mock_is_world_process_zero = mock_patches_for_evaluate["mock_is_world_process_zero"]

    # Mock functions that evaluate() calls.
    mock_cuda_is_available.return_value = True
    mock_is_image_text_llm.return_value = False
    mock_is_world_process_zero.return_value = True
    mock_get_task_dict.return_value = MagicMock()
    mock_generate_lm_harness_model_args.return_value = MagicMock()
    mock_lm_harness_get_model_class.return_value = MagicMock()

    with pytest.raises(
        ValueError,
        match=f"Unsupported inference engine type: {unsupported_inference_engine_type}."
        " Our integration with the `lm_harness` evaluation backend supports "
        "the `NATIVE`, `VLLM` and `REMOTE` inference_engine types.",
    ):
        evaluate_lm_harness(
            task_params=LMHarnessTaskParams(
                evaluation_backend="lm_harness",
                task_name="mmlu",
            ),
            config=EvaluationConfig(
                tasks=[],
                inference_engine=unsupported_inference_engine_type,
            ),
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any

from oumi.core.configs.params.synthesis_params import GeneratedAttribute
from oumi.core.synthesis.attribute_synthesizer import AttributeSynthesizer
from oumi.utils.logging import logger


class DataSynthesizer:
    """Synthesizes data using attributes from the dataset plan."""

    def __init__(
        self,
        generated_attributes: list[GeneratedAttribute],
        attribute_synthesizer: AttributeSynthesizer,
    ):
        """Initialize the synthesizer."""
        self._generated_attributes = generated_attributes
        self._attribute_synthesizer = attribute_synthesizer

    def synthesize(
        self,
        dataset_plan_samples: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Synthesize data using attributes from the dataset plan.

        If there are multiple generated attributes, each will be synthesized
        sequentially, with their resulting output added to the dataset plan records.
        """
        if not dataset_plan_samples:
            return dataset_plan_samples

        for generated_attribute in self._generated_attributes:
            logger.info(f"Synthesizing generated attribute: {generated_attribute.id}")
            synthesized_attribute_records = self._attribute_synthesizer.synthesize(
                dataset_plan_samples,
                generated_attribute,
            )
            for i, record in enumerate(synthesized_attribute_records):
                dataset_plan_samples[i].update(record)

        return dataset_plan_samples
